/*****************************************************************************
 *
 * DictD++, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#include "stdafx.h"

#include <base64.h>
#include <memory_mapped_file.h>

#include <unicode/ucnv.h>

#include "resort_utils.h"

//------------------------------------------------------------------------------
// IndexResorter implementation:
//------------------------------------------------------------------------------
IndexResorter::IndexResorter(const std::string&i_locale)
  {
  UErrorCode status = U_ZERO_ERROR;
  mp_collator = Collator::createInstance(i_locale.c_str(), status);

  if (U_FAILURE(status))
    {
    g_log << CRIT <<"Failed to create the collator" << endlog;
    exit(1);
    }

  mp_collator -> setAttribute(UCOL_STRENGTH, UCOL_IDENTICAL, status);

  if (U_FAILURE(status))
    {
    g_log << CRIT <<"Failed to set collator attributes" << endlog;
    exit(1);
    }
  }
//------------------------------------------------------------------------------
IndexResorter::~IndexResorter()
  {
  delete mp_collator;
  }
//------------------------------------------------------------------------------
void IndexResorter::AddKeyword(const UnicodeString& i_keyword, index_type i_start,
                               index_type i_length)
  {
  UnicodeString keyword_norm;
  UErrorCode status = U_ZERO_ERROR;

  Normalizer::normalize(i_keyword, UNORM_NFD, 0, keyword_norm, status);

  if (U_FAILURE(status))
    {
    g_log << CRIT << "Cannot normalize keyword" << endlog;
    exit(1);
    }

  CollationKey key;
  mp_collator -> getCollationKey(keyword_norm, key, status);

  if (U_FAILURE(status))
    {
    g_log << CRIT << "Failed to create the collation key" << endlog;
    exit(1);
    }

  m_info.push_back(KeyInfo(keyword_norm, i_start, i_length));
  m_keys.push_back(SortStructure(key, &m_info.back()));
  }
//------------------------------------------------------------------------------
void IndexResorter::SortAndWrite(std::ostream& os, const std::string& i_out_enc)
  {
  g_log << INFO << (ulong)m_keys.size() << " keywords loaded" << endlog;
  g_log << INFO << "Starting to sort" << endlog;

  std::vector<const SortStructure*> keys_refs;
  keys_refs.reserve(m_keys.size());

  for (std::deque<SortStructure>::const_iterator it = m_keys.begin(); it != m_keys.end(); ++it)
    keys_refs.push_back(&*it);

  // Let the fun begin! ;-)
  std::sort(keys_refs.begin(), keys_refs.end(), SortStructureRefsComparator());

  g_log << INFO << "Sorting has been finished" << endlog;
  g_log << INFO << "Starting to produce result index" << endlog;

  UErrorCode status = U_ZERO_ERROR;
  UConverter* p_res_converter = ucnv_open(i_out_enc.c_str(), &status);

  if (U_FAILURE(status))
    {
    g_log << CRIT << "Failed to create the converter" << endlog;
    exit(1);
    }

  for (std::vector<const SortStructure*>::const_iterator it = keys_refs.begin(); it != keys_refs.end(); ++it)
    {
    int32_t req_len = (*it) -> mp_info -> m_keyword.extract(NULL, 0, p_res_converter, status);
    char* p_buf = new char[req_len + 1];
    status = U_ZERO_ERROR;
    (*it) -> mp_info -> m_keyword.extract(p_buf, req_len + 1, p_res_converter, status);

    if (U_FAILURE(status))
      {
      g_log << CRIT << "Failed to convert keyword" << endlog;
      exit(1);
      }

    os << p_buf << "\t" << Tools::b64_encode((*it) -> mp_info -> m_start) << "\t" << Tools::b64_encode((*it) -> mp_info -> m_length) << "\n";

    delete[] p_buf;
    }

  g_log << INFO << "Result index is ready" << endlog;
  ucnv_close(p_res_converter);
  }
//------------------------------------------------------------------------------
 // Misc utility functions:
//------------------------------------------------------------------------------
void IndexFromXML(const std::string& i_dict_xml_name, const std::string& i_input_enc,
                  const std::string& i_res_index_name,
                  const std::string& i_locale, const std::string& i_output_enc)
  {
  MemoryMappedFile* p_map_dict_file = new MemoryMappedFile(i_dict_xml_name);
  IndexResorter resorter(i_locale);
  const char* p_dict = p_map_dict_file -> Get();
  index_type len = p_map_dict_file -> Length();
  std::string inp_enc;
  UConverter* p_converter = NULL;
  UnicodeString keyword;

  g_log << INFO << "Loading keywords" << endlog;
  index_type step = len/50, cur_pos = 0, saved_pos = 0;

  for (ulong pos = 0; pos < len;)
    {
    cur_pos += pos - saved_pos;
    saved_pos = pos;
    while (cur_pos > step)
      {
      cur_pos -= step;
      std::clog << "=";
      }

    if (p_dict[pos] == '<')
      {
      ulong old_pos = pos;
      for (++pos; p_dict[pos] != ' ' && p_dict[pos] != '>'; ++pos);
      std::string element(p_dict + old_pos + 1, pos - old_pos - 1);

      if (element == "?xml")
        {
        if (i_input_enc.empty())
          {
          old_pos = pos;
          for (++pos; p_dict[pos] != '>'; ++pos);
          std::string args(p_dict + old_pos, pos - old_pos);
          if (args.find("encoding") == std::string::npos)
            {
            g_log << CRIT << "\"encoding\" attribute was not found in \"?xml\" element" << endlog;
            exit(1);
            }
          std::string::size_type enc_pos = args.find("encoding");
          enc_pos += 10;
          std::string::size_type enc_old_pos = enc_pos;
          while (args[enc_pos] != '\"') ++enc_pos;
          inp_enc = args.substr(enc_old_pos, enc_pos - enc_old_pos);
          g_log << INFO << "Using input encoding \"" << inp_enc << "\" found in \"?xml\" section" << endlog;
          }
        }
      else if (element == "key")
        {
        old_pos = pos + 1;
        while (p_dict[++pos] != '<');
        if (!p_converter)
          {
          UErrorCode status = U_ZERO_ERROR;
          if (i_input_enc.size())
            p_converter = ucnv_open(i_input_enc.c_str(), &status);
          else if (inp_enc.size())
            p_converter = ucnv_open(inp_enc.c_str(), &status);
          else
            {
            g_log << CRIT << "Unknown input encoding" << std::endl;
            exit(1);
            }

          if (U_FAILURE(status))
            {
            g_log << CRIT << "Failed to create the converter" << endlog;
            exit(1);
            }
          }

        UErrorCode status = U_ZERO_ERROR;
        keyword = UnicodeString(p_dict + old_pos, pos - old_pos, p_converter, status);
        if (U_FAILURE(status))
          {
          g_log << CRIT << "Failed to convert the keyword" << endlog;
          exit(1);
          }
        }
      else if (element == "def")
        {
        while (pos < len && p_dict[pos] != '\r' && p_dict[pos] != '\n')
          ++pos;
        keyword.trim();
        resorter.AddKeyword(keyword, old_pos, pos - old_pos);
        }
      }
    while (pos < len && p_dict[pos] != '\r' && p_dict[pos] != '\n')
      ++pos;
    if (pos < len)
      ++pos;
    if (pos < len && p_dict[pos] == '\n')
      ++pos;
    }
  std::clog << std::endl;

  delete p_map_dict_file;
  ucnv_close(p_converter);
  std::ofstream ofs(i_res_index_name.c_str(), std::ios::out | std::ios::binary);
  resorter.SortAndWrite(ofs, i_output_enc);
  }
//------------------------------------------------------------------------------
// This function will read input index in "input" encoding and create resulting index
// that is sorted according to given locale and has "output" encoding
void ResortIndex(const std::string& i_index_name, const std::string& i_input_enc,
                 const std::string& i_res_index_name, const std::string& i_locale,
                 const std::string& i_output_enc, bool i_fl_numeric)
  {
  MemoryMappedFile* p_map_index_file = new MemoryMappedFile(i_index_name);
  IndexResorter resorter(i_locale);

  const char* p_index = p_map_index_file -> Get();
  index_type index_len = p_map_index_file -> Length();

  UErrorCode status = U_ZERO_ERROR;
  UConverter* p_converter = ucnv_open(i_input_enc.c_str(), &status);

  if (U_FAILURE(status))
    {
    g_log << CRIT << "Failed to create the converter" << endlog;
    exit(1);
    }

  index_type step = index_len/50, cur_pos = 0, saved_pos = 0;

  for (index_type pos = 0; pos < index_len; )
    {
    cur_pos += pos - saved_pos;
    saved_pos = pos;
    while (cur_pos > step)
      {
      cur_pos -= step;
      std::clog << "=";
      }

    index_type old_pos = pos;

    while (pos < index_len && p_index[pos] != '\t' && p_index[pos] != '\n')
      ++pos;
      
    if (pos >= index_len || p_index[pos] != '\t')
      {
      g_log << CRIT << std::string("Wrong format of index file ") +
                       p_map_index_file -> GetName() +
                       std::string(" at offset ") + boost::lexical_cast<std::string>(pos) << endlog;

      exit(1);
      }

    //g_log << CRIT << "Keyword: " << std::string(p_index + old_pos, pos - old_pos) << endlog;

    UErrorCode status = U_ZERO_ERROR;
    UnicodeString keyword(p_index + old_pos, pos - old_pos, p_converter, status);

    if (U_FAILURE(status))
      {
      g_log << CRIT << "Failed to convert keyword" << endlog;
      exit(1);
      }

    keyword.trim();

    old_pos = ++pos; // we are now at the start of 'start' value

    while (pos < index_len && p_index[pos] != '\t' && p_index[pos] != '\n')
      ++pos;

    if (p_index[pos] != '\t')
      {
      g_log << CRIT << "Incorrect index format" << endlog;
      exit(1);
      }

    index_type pos1 = pos;
    ++pos;

    while (pos < index_len && p_index[pos] != '\t' && p_index[pos] != '\r' && p_index[pos] != '\n')
      ++pos;

    if (p_index[pos] != '\r' && p_index[pos] != '\n')
      {
      g_log << CRIT << "Incorrect index format" << endlog;
      exit(1);
      }

    index_type start, len;

    if (i_fl_numeric)
      {
      start = boost::lexical_cast<index_type>(std::string(p_index + old_pos, p_index + pos1));
      len = boost::lexical_cast<index_type>(std::string(p_index + pos1 + 1, p_index + pos));
      }
    else
      {
      Tools::b64_decode(start, p_index + old_pos, pos1 - old_pos);
      Tools::b64_decode(len, p_index + pos1 + 1, pos - pos1 - 1);
      }

    //g_log << CRIT << "Start: " << start << ", len: " << len << endlog;

    resorter.AddKeyword(keyword, start, len);

    if (pos < index_len && p_index[pos] == '\r') // skip trailing newline symbol
      ++pos;
    if (pos < index_len && p_index[pos] == '\n') // skip trailing newline symbol
      ++pos;
    }

  std::clog << std::endl;

  delete p_map_index_file;

  std::ofstream ofs(i_res_index_name.c_str(), std::ios::out | std::ios::binary);
  resorter.SortAndWrite(ofs, i_output_enc);

  ucnv_close(p_converter);
  }
//------------------------------------------------------------------------------
void IndexFromHD(const std::string& i_dict_name, const std::string& i_input_enc,
                 const std::string& i_out_index_name, const std::string& i_locale,
                 const std::string& i_output_enc)
  {
  MemoryMappedFile* p_mem_file = new MemoryMappedFile(i_dict_name);
  IndexResorter resorter(i_locale);

  unsigned const char* p_dict = (unsigned const char*)p_mem_file -> Get();
  index_type len = p_mem_file -> Length();

  index_type num = 0;

  UErrorCode status = U_ZERO_ERROR;
  UConverter* p_converter = ucnv_open(i_input_enc.c_str(), &status);

  index_type step = len/50, cur_pos = 0, saved_pos = 0;

  for (index_type pos = 0; pos < len;)
    {
    cur_pos += pos - saved_pos;
    saved_pos = pos;
    while (cur_pos > step)
      {
      cur_pos -= step;
      std::clog << "=";
      }

    if (p_dict[pos] != '%' && p_dict[pos+1] != 'h')
      {
      g_log << ERR << "%h not found at position " << pos << ", terminating index construction" << endlog;
      return;
      }

    for (pos += 2; pos < len && p_dict[pos] <= ' '; ++pos);

    if (pos == len)
      {
      g_log << ERR << "Format error at position " << pos << ", terminating index construction" << endlog;
      return;
      }

    index_type old_pos = pos;

    for (; pos < len && p_dict[pos] >= ' '; ++pos);

    if (pos == len || pos == old_pos)
      {
      g_log << ERR << "Format error at position " << pos << ", terminating index construction" << endlog;
      return;
      }
    
    UErrorCode status = U_ZERO_ERROR;
    UnicodeString keyword((const char*)p_dict + old_pos, pos - old_pos, p_converter, status);
    if (U_FAILURE(status))
      {
      g_log << CRIT << "Cannot convert keyword" << endlog;
      exit(1);
      }

    keyword.trim();

    for (; pos < len && p_dict[pos] <= ' '; ++pos);

    if (p_dict[pos] != '%' && p_dict[pos+1] != 'd')
      {
      g_log << ERR << "%d not found at position " << pos << ", terminating index construction" << endlog;
      return;
      }

    for (pos += 2; pos < len && p_dict[pos] <= ' '; ++pos);

    if (pos == len)
      {
      g_log << ERR << "Format error at position " << pos << ", terminating index construction" << endlog;
      return;
      }

    old_pos = pos;

    for (;pos < len && (p_dict[pos] != '%' || p_dict[pos+1] != 'h'); ++pos);

    resorter.AddKeyword(keyword, old_pos, pos - old_pos);
    ++num;
    }

  std::clog << std::endl;

  delete p_mem_file;
  ucnv_close(p_converter);

  std::ofstream ofs(i_out_index_name.c_str(), std::ios::out | std::ios::binary);
  resorter.SortAndWrite(ofs, i_output_enc);
  }
//------------------------------------------------------------------------------