/*****************************************************************************
 *
 * DictD++, Copyright (C) 2007 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#include "dictclient++.h"

#include <iostream>
#include <boost/regex.hpp>
#include <boost/tokenizer.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/regex.hpp>
#include <boost/asio.hpp>

#include "md5.hpp"

//-----------------------------------------------------------------------------
struct DictClient::DictClientImpl
  {
  boost::asio::io_service m_io_service;
  boost::asio::ip::tcp::socket m_socket;
  std::vector<std::string> m_capabilities;
  std::string m_msg_id;

  DictClientImpl(): m_socket(m_io_service) {}
  };
//-----------------------------------------------------------------------------
const std::string client_version = "0.0.1";
//-----------------------------------------------------------------------------
using boost::asio::ip::tcp;
//-----------------------------------------------------------------------------
DictClient::DictClient(const std::string& server_name, const std::string& service_name):
  mp_impl(new DictClientImpl())
  {
  // Get a list of endpoints corresponding to the server name.
  tcp::resolver resolver(mp_impl -> m_io_service);
  tcp::resolver::query query(server_name, service_name);
  tcp::resolver::iterator endpoint_it = resolver.resolve(query);
  tcp::resolver::iterator end;

  // Try each endpoint until we successfully establish a connection.
  boost::asio::error error = boost::asio::error::host_not_found;
  while (error && endpoint_it != end)
    {
    mp_impl -> m_socket.close();
    mp_impl -> m_socket.connect(*endpoint_it++, boost::asio::assign_error(error));
    }

  if (error)
    throw error;

  boost::asio::streambuf resp;
  boost::asio::read_until(mp_impl -> m_socket, resp, boost::regex("\r\n"));

  // Analyse server response
  std::istream resp_s(&resp);
  unsigned status;
  resp_s >> status;
  std::string status_msg;
  std::getline(resp_s, status_msg);

  if (status != 220) // normal server greeting code
    throw std::runtime_error("Invalid server response code: " + boost::lexical_cast<std::string>(status));

  // Hm, why capabilities field is completely optional rather than just allowing empty capabilities? Strange ...
  boost::match_results<std::string::const_iterator> what;
  if (boost::regex_search(status_msg, what, boost::regex("<(.*)>\\s*(<.*>)"), boost::match_default))
    {
    std::string capa = what[1];
    typedef boost::tokenizer<boost::char_separator<char> > tokenizer;
    tokenizer tokens(capa, boost::char_separator<char>("."));

    for (tokenizer::iterator tok_iter = tokens.begin(); tok_iter != tokens.end(); ++tok_iter)
      mp_impl -> m_capabilities.push_back(*tok_iter);

    mp_impl -> m_msg_id = what[2];
    }
  else if (boost::regex_search(status_msg, what, boost::regex("<(.*)>"), boost::match_default))
    mp_impl -> m_msg_id = what[1];
  else
    throw std::runtime_error("Invalid server greeting");

  boost::asio::streambuf req;
  std::ostream req_s(&req);
  req_s << "CLIENT DICT++/" << client_version << "\r\n";
  boost::asio::write(mp_impl -> m_socket, req);

  boost::asio::read_until(mp_impl -> m_socket, resp, boost::regex("\r\n"));

  // Check that response is OK.
  resp_s >> status;

  if (status != 250)
    throw std::runtime_error("CLIENT command failed, status code: " + boost::lexical_cast<std::string>(status));
  }
//-----------------------------------------------------------------------------
DictClient::~DictClient()
  {
  boost::asio::streambuf req;
  std::ostream req_s(&req);
  req_s << "QUIT\r\n";
  boost::asio::write(mp_impl -> m_socket, req);

  boost::asio::streambuf resp;
  std::istream resp_s(&resp);

  boost::asio::read_until(mp_impl -> m_socket, resp, boost::regex("\r\n"));

  // Check that response is OK.
  unsigned status;
  resp_s >> status;

  if (status != 221) // Actually, we do not care too much because the session is finished
    std::cerr << "QUIT command failed, status code: " + boost::lexical_cast<std::string>(status) << std::endl;

  delete mp_impl;
  }
//-----------------------------------------------------------------------------
void DictClient::Auth(const std::string& user_name, const std::string& password)
  {
  boost::asio::streambuf req;
  std::ostream req_s(&req);

  req_s << "AUTH " << user_name << " " << boost::md5((mp_impl -> m_msg_id + password).c_str()).digest().hex_str_value() << "\r\n";
  boost::asio::write(mp_impl -> m_socket, req);

  boost::asio::streambuf resp;
  std::istream resp_s(&resp);

  boost::asio::read_until(mp_impl -> m_socket, resp, boost::regex("\r\n"));

  // Check that response is OK.
  unsigned status;
  resp_s >> status;

  if (status != 230)
    throw std::runtime_error("AUTH command failed, status code: " + boost::lexical_cast<std::string>(status));
  }
//-----------------------------------------------------------------------------
void DictClient::ProcessCommandCommon(PairItems& items, const std::string& command, unsigned not_found_code, unsigned success_code)
  {
  boost::asio::streambuf req;
  std::ostream req_s(&req);
  req_s << command << "\r\n";
  boost::asio::write(mp_impl -> m_socket, req);

  boost::asio::streambuf resp;
  std::istream resp_s(&resp);

  boost::asio::read_until(mp_impl -> m_socket, resp, boost::regex("\r\n"));

  // Check that response is OK.
  unsigned status;
  resp_s >> status;

  if (status == not_found_code)
    return;

  if (status != success_code)
    throw std::runtime_error(command + " command failed, status code: " + boost::lexical_cast<std::string>(status));

  resp_s.ignore(std::numeric_limits<std::streamsize>::max(), '\n'); // Skip the rest of the line

  boost::asio::read_until(mp_impl -> m_socket, resp, boost::regex("\r\n.\r\n"));

  std::string line;
  while (std::getline(resp_s, line) && line != ".\r")
    {
    std::string::size_type ind = line.find_last_of('\r');
    if (ind != std::string::npos)
      line.erase(ind);

    typedef boost::tokenizer<boost::escaped_list_separator<char> > tokenizer;
    tokenizer tokens(line, boost::escaped_list_separator<char>("\\", " \t", "\"\'"));
    
    PairItem item;
    tokenizer::iterator tok_iter = tokens.begin();
    if (tok_iter == tokens.end())
      throw std::runtime_error("Invalid " + command + " command response");
    item.first = *tok_iter; ++tok_iter;
    if (tok_iter == tokens.end())
      throw std::runtime_error("Invalid " + command + " command response");
    item.second = *tok_iter; ++tok_iter;
    if (tok_iter != tokens.end())
      throw std::runtime_error("Invalid " + command + " command response");
     
    items.push_back(item);
    }

  resp_s >> status;
  if (status != 250)
    throw std::runtime_error(command + " command failed, status code: " + boost::lexical_cast<std::string>(status));
  }
//-----------------------------------------------------------------------------
void DictClient::GetDBs(DBs& dbs)
  {
  ProcessCommandCommon(dbs, "SHOW DB", 554, 110);
  }
//-----------------------------------------------------------------------------
void DictClient::GetStrats(Strats& strats)
  {
  ProcessCommandCommon(strats, "SHOW STRAT", 555, 111);
  }
//-----------------------------------------------------------------------------
void DictClient::Match(Matches& matches, const std::string& db_name,
                       const std::string& strat_name, const std::string& word)
  {
  boost::asio::streambuf req;
  std::ostream req_s(&req);

  ProcessCommandCommon(matches, "MATCH " + db_name + " " + strat_name + " " + word, 552, 152);
  }
//-----------------------------------------------------------------------------
void DictClient::Define(Definitions& definitions, const std::string& db_name,
                        const std::string& word)
  {
  boost::asio::streambuf req;
  std::ostream req_s(&req);
  req_s << "DEFINE " << db_name << " " << word << "\r\n";

  boost::asio::write(mp_impl -> m_socket, req);

  boost::asio::streambuf resp;
  boost::asio::read_until(mp_impl -> m_socket, resp, boost::regex("\r\n"));

  std::istream resp_s(&resp);
  unsigned status;

  // Check that response is OK.
  resp_s >> status;

  if (status == 552) // no matches
    return;

  if (status != 150)
    throw std::runtime_error("DEFINE command failed, status code: " + boost::lexical_cast<std::string>(status));

  resp_s.ignore(std::numeric_limits<std::streamsize>::max(), '\n'); // Skip the rest of the line

  while (true)
    {
    boost::asio::read_until(mp_impl -> m_socket, resp, boost::regex("\r\n"));
    resp_s >> status;

    if (status == 250) // We've finished definitions parsing
      break;

    if (status != 151)
      throw std::runtime_error("DEFINE command failed, status code: " + boost::lexical_cast<std::string>(status));

    std::string line;
    std::getline(resp_s, line);
    std::string::size_type ind = line.find_last_of('\r');
    if (ind != std::string::npos)
      line.erase(ind);

    ind = line.find_first_not_of(' ');
    if (ind == std::string::npos)
      throw std::runtime_error("Invalid DEFINE command response");

    typedef boost::tokenizer<boost::escaped_list_separator<char> > tokenizer;
    tokenizer tokens(line.begin() + ind, line.end(), boost::escaped_list_separator<char>("\\", " \t", "\"\'"));
    
    definitions.push_back(DefinitionItem());
    DefinitionItem& item = definitions.back();
    tokenizer::iterator tok_iter = tokens.begin();
    if (tok_iter == tokens.end())
      throw std::runtime_error("Invalid DEFINE command response");
    item.m_word = *tok_iter; ++tok_iter;
    if (tok_iter == tokens.end())
      throw std::runtime_error("Invalid DEFINE command response");
    item.m_dict_name = *tok_iter; ++tok_iter;
    if (tok_iter == tokens.end())
      throw std::runtime_error("Invalid DEFINE command response");
    item.m_dict_descr = *tok_iter; ++tok_iter;
    // tok_iter may point not to the end() here because there could be other textual information,
    // such as "text follows" or whatever
     
    std::size_t size = boost::asio::read_until(mp_impl -> m_socket, resp, boost::regex("\r\n.\r\n"));
    std::vector<char> vec(size - 5); // we should skip the trailing 5 chars as they're "\r\n.\r\n"
    resp_s.read(&vec[0], size - 5);
    item.m_definition.assign(vec.begin(), vec.end());

    resp_s.ignore(5, 0);
    }
  }
//-----------------------------------------------------------------------------
