/*****************************************************************************
 *
 * DictD++, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#include "stdafx.h"
#include "abstract_index.h"
#include "abstract_memory_mapper.h"
#include "dictd_index_cacher.h"
#include "base64.h"
#include <unicode/ucnv.h>

//------------------------------------------------------------------------------
const char* DictdIndexCacher::g_cache_file_basic_sig = "DictD++ cache file (basic)";
const char* DictdIndexCacher::g_cache_file_ext_sig = "DictD++ cache file (extended)";
//------------------------------------------------------------------------------
// DictdCacheInfoBasic implementation
//------------------------------------------------------------------------------
void DictdCacheInfoBasic::Write(std::ostream& os,
				index_type i_keyword_pos, int32_t i_keyword_len,
				storage_type i_def_start, storage_type i_def_len,
				const CollationKey& i_key)
  {
  // keyword position in the index
  os.write((char*)&i_keyword_pos, sizeof(i_keyword_pos));
  // keyword length (in index character units)
  os.write((char*)&i_keyword_len, sizeof(i_keyword_len));
  // definition position in the storage
  os.write((char*)&i_def_start, sizeof(i_def_start));
  // definition length
  os.write((char*)&i_def_len, sizeof(i_def_len));

  int32_t key_size;
  const uint8_t* p_key = i_key.getByteArray(key_size);
  // collation key length
  os.write((char*)&key_size, sizeof(key_size));
  // collation key
  os.write((char*)p_key, key_size);
  }
//------------------------------------------------------------------------------
// DictdCacheInfoExtended implementation
//------------------------------------------------------------------------------
void DictdCacheInfoExtended::Write(std::ostream& os,
				   storage_type i_def_start, storage_type i_def_len,
				   const CollationKey& i_key, const UnicodeString& i_keyword)
  {
  // definition position in the storage
  os.write((char*)&i_def_start, sizeof(i_def_start));
  // definition length
  os.write((char*)&i_def_len, sizeof(i_def_len));

  int32_t key_size;
  const uint8_t* p_key = i_key.getByteArray(key_size);
  // collation key length
  os.write((char*)&key_size, sizeof(key_size));

  int32_t keyword_len = i_keyword.length();
  // keyword length
  os.write((char*)&keyword_len, sizeof(keyword_len));
  // collation key
  os.write((char*)p_key, key_size*sizeof(uint8_t));
  // keyword in UTF-16
  os.write((char*)i_keyword.getBuffer(), i_keyword.length()*sizeof(UChar));
  }
//------------------------------------------------------------------------------
// DictdIndexCacher implementation:
//------------------------------------------------------------------------------
void DictdIndexCacher::Initialize() throw(IndexException)
  {
  UErrorCode status = U_ZERO_ERROR;
  mp_collator = Collator::createInstance(m_locale.c_str(), status);

  if (U_FAILURE(status))
    throw IndexException("CollatorCreationFailed", std::string("Failed to create the collator for locale ") + m_locale);

  mp_collator -> setStrength(Collator::PRIMARY);

  mp_converter = ucnv_open(m_codepage.c_str(), &status);

  if (U_FAILURE(status))
    throw IndexException("ConverterCreationFailed", std::string("Failed to create the converter for codepage ") + m_codepage);
  }
//------------------------------------------------------------------------------
void DictdIndexCacher::DeInitialize()
  {
  delete mp_collator;

  if (mp_converter)
    ucnv_close(mp_converter);
  }
//------------------------------------------------------------------------------
DictdIndexCacher::DictdIndexCacher(const std::string& i_locale, const std::string& i_codepage,
				   const AbstractMemoryMapper& i_index_file,
				   const std::string& i_index_id, bool i_fl_extended):
		    m_locale(i_locale), m_codepage(i_codepage), m_index_file(i_index_file),
		    m_index_id(i_index_id), m_fl_extended(i_fl_extended),
		    mp_collator(NULL), mp_converter(NULL)
  {
  Initialize();
  }
//------------------------------------------------------------------------------
DictdIndexCacher::~DictdIndexCacher() throw()
  {
  DeInitialize();
  }
//------------------------------------------------------------------------------
bool DictdIndexCacher::IsCacheValid(std::iostream& io_cache) const
  {
  g_log << DEBUG << "Cache file info follows" << endlog;

  std::string str;
  std::getline(io_cache, str);
  g_log << DEBUG << str << endlog;
  if (str != g_cache_file_basic_sig &&
      str != g_cache_file_ext_sig)
    return false;

  if (m_fl_extended != (str == g_cache_file_ext_sig))
    return false;

  std::getline(io_cache, str);
  g_log << DEBUG << str << endlog;
  if (str != VERSION_STR)
    return false;

  std::getline(io_cache, str);
  g_log << DEBUG << str << endlog;
  if (str != m_index_id)
    return false;

  std::getline(io_cache, str);
  g_log << DEBUG << str << endlog;
  if (str != m_locale)
    return false;

  std::getline(io_cache, str);
  g_log << DEBUG << str << endlog;
  if (str != m_codepage)
    return false;

  UVersionInfo ver_info, cur_ver_info;
  mp_collator -> getVersion(cur_ver_info);

  io_cache.read((char*)&ver_info, U_MAX_VERSION_LENGTH); // Collator version info  
  for (ulong i = 0; i < U_MAX_VERSION_LENGTH; ++i)
    if (ver_info[i] != cur_ver_info[i])
      return false;

  g_log << DEBUG << "Cache file is valid" << endlog;
  return true;
  }
//------------------------------------------------------------------------------
void DictdIndexCacher::GetIndexEntry(storage_type& o_start, storage_type& o_length,
				     index_type& o_keyword_length, UnicodeString& o_keyword,
				     const char*& p_cur_pos) const throw(IndexException)
  {
  const char* p_index = m_index_file.Get();
  const char* p_end_pos = p_index + m_index_file.Length();
  UErrorCode status = U_ZERO_ERROR;

  const char* p_start_pos = p_cur_pos, *p_old_pos;
  UChar32 c = ucnv_getNextUChar(mp_converter, &p_cur_pos, p_end_pos, &status);

  assert (c != '\n');

  o_keyword.remove();

  while (c != '\t' && c != '\n' && p_cur_pos < p_end_pos)
    {
    o_keyword += c;
    p_old_pos = p_cur_pos;
    c = ucnv_getNextUChar(mp_converter, &p_cur_pos, p_end_pos, &status);
    }

  o_keyword_length = (index_type)(p_old_pos - p_start_pos);

  if (p_cur_pos >= p_end_pos || c != '\t')
    throw IndexException("WrongDictdIndexFile", std::string("Wrong format of index file ") +
			 m_index_file.GetName() +
			 std::string(" at offset ") + boost::lexical_cast<std::string>((index_type)(p_cur_pos - p_index)));
    
  UnicodeString str_pos, str_length;

  c = ucnv_getNextUChar(mp_converter, &p_cur_pos, p_end_pos, &status);
  while (c != '\t' && c != '\n' && p_cur_pos < p_end_pos)
    {
    str_pos += c;
    c = ucnv_getNextUChar(mp_converter, &p_cur_pos, p_end_pos, &status);
    }
      
  if (p_cur_pos >= p_end_pos || c != '\t')
    throw IndexException("WrongDictdIndexFile", std::string("Wrong format of index file ") +
			 m_index_file.GetName() +
			 std::string(" at offset ") + boost::lexical_cast<std::string>((index_type)(p_cur_pos - p_index)));

  c = ucnv_getNextUChar(mp_converter, &p_cur_pos, p_end_pos, &status);
  while (c != '\n' && p_cur_pos < p_end_pos)
    {
    str_length += c;
    c = ucnv_getNextUChar(mp_converter, &p_cur_pos, p_end_pos, &status);
    }
      
  if (p_cur_pos > p_end_pos || c != '\n')
    throw IndexException("WrongDictdIndexFile", std::string("Wrong format of index file ") +
			 m_index_file.GetName() +
			 std::string(" at offset ") + boost::lexical_cast<std::string>((index_type)(p_cur_pos - p_index)));

  std::string str_pos1(AsStringTrunc(str_pos));
  std::string str_length1(AsStringTrunc(str_length));

  if (!Tools::b64_decode(o_start, str_pos1.c_str(), (ulong)str_pos1.length()))
    throw IndexException("WrongDictdIndexFile", std::string("Wrong format of index file ") +
			 m_index_file.GetName() +
			 std::string(": bad base64 at offset ") + boost::lexical_cast<std::string>((index_type)(p_cur_pos - p_index)));
    
  if (!Tools::b64_decode(o_length, str_length1.c_str(), (ulong)str_length1.length()))
    throw IndexException("WrongDictdIndexFile", std::string("Wrong format of index file ") +
			 m_index_file.GetName() +
			 std::string(": bad base64 at offset ") + boost::lexical_cast<std::string>((index_type)(p_cur_pos - p_index)));

  if (U_FAILURE(status))
    throw IndexException("WrongDictdIndexFile", std::string("Wrong format of index file ") +
			 m_index_file.GetName() +
			 std::string(": cannot convert to Unicode, end line pos is ") + boost::lexical_cast<std::string>((index_type)(p_cur_pos - p_index)));
  }
//------------------------------------------------------------------------------
void DictdIndexCacher::CreateCache(std::iostream& io_cache) const
  {
  g_log << INFO << "Rebuilding cache for the file " << m_index_file.GetName() << endlog;

  UVersionInfo ver_info;
  mp_collator -> getVersion(ver_info);

  io_cache.clear();
  io_cache.seekg(0, std::ios::beg);

  if (m_fl_extended)
    io_cache << g_cache_file_ext_sig << std::endl;
  else
    io_cache << g_cache_file_basic_sig << std::endl;

  io_cache << VERSION_STR << std::endl; // DictD++ server version

  io_cache << m_index_id << std::endl; // Some index info like it's size/date, etc

  io_cache << m_locale << std::endl; // Locale name used to create collation keys

  io_cache << m_codepage << std::endl; // Codepage used to decode index
  
  io_cache.write((char*)&ver_info, U_MAX_VERSION_LENGTH); // Collator version info

  index_type keywords_num = 0;

  // We are going to create cache entry for each index item.

  const char *p_start_pos = m_index_file.Get(), *p_cur_pos = p_start_pos,
	     *p_end_pos = p_start_pos + m_index_file.Length();

  std::deque<index_type> cache_pos;

  CollationKey last_key;
  UnicodeString keyword;
  index_type keyword_length;
  storage_type start, length;

  while (p_cur_pos < p_end_pos)
    {
    const char* p_old_pos = p_cur_pos;
    cache_pos.push_back(io_cache.tellg());
    GetIndexEntry(start, length, keyword_length, keyword, p_cur_pos);
      
    UErrorCode status = U_ZERO_ERROR;
    CollationKey key;
    mp_collator -> getCollationKey(keyword, key, status);

    if (U_FAILURE(status))
      throw IndexException("CollationKeyCreationFailed", "Failed to create the collation key in DictdIndex cache construction");

    if (m_fl_extended)
      DictdCacheInfoExtended::Write(io_cache, start, length, key, keyword);
    else
      DictdCacheInfoBasic::Write(io_cache, (index_type)(p_old_pos - p_start_pos), keyword_length,
				 start, length, key);

    ++keywords_num;

    if (last_key.compareTo(key) > Collator::EQUAL)
      throw IndexException("WrongKeywordsOrder", "Keywords order in index is wrong");

    last_key = key;
    }

  index_type cur_cache_pos = io_cache.tellg();

  for (std::deque<index_type>::const_iterator it = cache_pos.begin(); it != cache_pos.end(); ++it)
    io_cache.write((char*)&*it, sizeof(index_type));

  g_log << TRACE << "Keywords number is " << keywords_num << endlog;

  io_cache.write((char*)&keywords_num, sizeof(keywords_num));
  io_cache.write((char*)&cur_cache_pos, sizeof(cur_cache_pos));
  }
//------------------------------------------------------------------------------
