/*****************************************************************************
 *
 * DictD++, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#include "stdafx.h"
#include "dictionary.h"
#include "abstract_index.h"
#include "abstract_storage.h"
#include "abstract_strategy.h"
#include "abstract_transformer.h"
#include "abstract_controller.h"
#include "access_rights.h"
#include "convertion.h"

//------------------------------------------------------------------------------
Dictionary::Dictionary(const UnicodeString& i_name, const UnicodeString& i_descr,
		       const std::string& i_def_out_mime_type, const AbstractIndex* ip_index,
		       const AbstractStorage* ip_storage, const std::list<Convertion*>* ip_convertions,
                       const AccessRights* ip_access_rights) throw(DictionaryException):

  m_name(i_name), m_descr(i_descr), m_def_out_mime_type(i_def_out_mime_type), mp_index(ip_index),
  mp_storage(ip_storage), mp_convertions(ip_convertions), mp_access_rights(ip_access_rights)
  {
  }
//------------------------------------------------------------------------------
Dictionary::~Dictionary() throw()
  {
  delete mp_index;
  delete mp_storage;

  for (std::list<Convertion*>::const_iterator it = mp_convertions -> begin(); it != mp_convertions -> end(); ++it)
    delete *it;

  delete mp_convertions;
  delete mp_access_rights;
  }
//------------------------------------------------------------------------------
void Dictionary::Validate() const throw(DictionaryException)
  {
  try
    {
    for (std::list<Convertion*>::const_iterator it = mp_convertions -> begin();
	 it != mp_convertions -> end(); ++it)
      (*it) -> Validate();
    }
  catch(const TransformerException&)
    {
    throw DictionaryException("ValidationFailed", "Was not able to validate convertions");
    }
  }
//------------------------------------------------------------------------------
UnicodeString Dictionary::GetDictionaryName() const throw()
  {
  return m_name;
  }
//------------------------------------------------------------------------------
UnicodeString Dictionary::GetDictionaryDescription() const throw()
  {
  return m_descr;
  }
//------------------------------------------------------------------------------
UnicodeString Dictionary::GetDictionaryStatistics() const throw()
  {
  return UnicodeString();
  }
//------------------------------------------------------------------------------
// TODO: I'm not really sure what should I do here - how can
// I know what are the limitations of the strategy implied by
// it's algorithm? (f.e. smth like soundex strategy in dictd ...)
// Should this be included into config file? Should I define some
// categories based on which it's possible to decide what strategy is
// applicable? Probably it's better to wait with this and see what
// strategies will be implemented and used, then it should be easier
// to find the most clean solution. $AT
bool Dictionary::IsStrategySupported(const AbstractStrategy& i_strategy) const throw()
  {
  return true;
  }
//------------------------------------------------------------------------------
std::string Dictionary::GetMimeType() const throw()
  {
  return mp_storage -> GetMimeType();
  }
//------------------------------------------------------------------------------
const AccessRights& Dictionary::GetAccessRights() const throw()
  {
  return *mp_access_rights;
  }
//------------------------------------------------------------------------------
void Dictionary::InitConvertions() throw(DictionaryException)
  {
  std::string mime_type = mp_storage -> GetMimeType();

  for (std::list<Convertion*>::const_iterator it = mp_convertions -> begin();
       it != mp_convertions -> end(); ++it)
    {
    std::string cur_in_mime_type = (*it) -> GetInputMimeType();
    std::string cur_out_mime_type = (*it) -> GetOutputMimeType();

    if (cur_in_mime_type != mime_type)
      throw DictionaryException("WrongConvertion", std::string("Convertion with input MIME type ") + cur_in_mime_type
						   + " is not applicable to the dictionary with MIME type " + mime_type);

    if (m_convertions.find(cur_out_mime_type) != m_convertions.end())
      throw DictionaryException("WrongConvertion", std::string("Convertion with output MIME type ") + cur_out_mime_type
						   + " already exists for this dictionary");

    m_convertions[cur_out_mime_type] = *it;
    }
  }
//------------------------------------------------------------------------------
void Dictionary::FindKeyWords(std::list<UnicodeString>& o_keywords, AbstractController& io_controller, const UnicodeString& i_word, const AbstractStrategy& i_strategy) const throw(DictionaryException)
  {
  try
    {
    std::list<index_type> keywords;
    i_strategy.FindKeyWords(keywords, io_controller, *mp_index, i_word);

    o_keywords.clear();

    for (std::list<index_type>::const_iterator it = keywords.begin(); it != keywords.end(); ++it)
      o_keywords.push_back(mp_index -> GetKeyWord(*it));
    }
  catch(const StrategyException& ex)
    {
    g_log << ex << endlog;
    throw DictionaryException("FindKeyWordsFail", "Failure during keywords searching");
    }
  }
//------------------------------------------------------------------------------
void Dictionary::FindDefinitions(Definitions& o_definitions, AbstractController& io_controller, const UnicodeString& i_word,
				 const AbstractStrategy& i_strategy, const std::list<MimeType>& i_mime_types,
                                 bool i_fl_show_mime_header) const throw(DictionaryException)
  {
  std::list<index_type> keywords;
  i_strategy.FindKeyWords(keywords, io_controller, *mp_index, i_word);
  std::string mime_type = ProduceOutputMimeType(i_mime_types);

  o_definitions.clear();

  for (std::list<index_type>::const_iterator it = keywords.begin(); it != keywords.end(); ++it)
    {
    if (!io_controller.IsDefineAllowed())
      break;
    io_controller.OneMoreDefine();

    AbstractPosition* p_pos = mp_index -> GetDefinitionPos(*it);
    
    UnicodeString res;

    if (i_fl_show_mime_header)
      {
      res = "Content-type: ";
      res += mime_type.c_str();
      res += "; charset=utf-8\r\nContent-transfer-encoding: 8bit\r\n\r\n";
      }

    if (mime_type == GetMimeType())
      res += mp_storage -> Get(*p_pos);
    else
      {
      if (m_convertions.find(mime_type) == m_convertions.end())
	{
	delete p_pos;
	throw DictionaryException("UnsupportedMimeType", std::string("This dictionary does not support convertion to MIME type ") + mime_type);
	}

      res += m_convertions.find(mime_type) -> second -> Apply(mp_storage -> Get(*p_pos));
      }

    o_definitions.push_back(DefinitionItem(mp_index -> GetKeyWord(*it), res));

    delete p_pos;
    }
  }
//------------------------------------------------------------------------------
std::string Dictionary::ProduceOutputMimeType(const std::list<MimeType>& i_types) const
  {
  for (std::list<MimeType>::const_iterator it = i_types.begin();
       it != i_types.end(); ++it)
    {
    if (it -> Matches(GetMimeType()))
      return GetMimeType();

    for (std::map<std::string, Convertion*>::const_iterator it1 = m_convertions.begin();
         it1 != m_convertions.end(); ++it1)
      if (it -> Matches(it1 -> first))
        return it1 -> first;
    }

  return "";
  }
//------------------------------------------------------------------------------
