/*****************************************************************************
 *
 * DictD++, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#include "stdafx.h"
#include "dictionary_list.h"
#include "dictionary.h"
#include "abstract_strategy.h"
#include "abstract_transformer.h"
#include "abstract_controller.h"
#include "access_rights.h"
#include "session_info.h"

//------------------------------------------------------------------------------
DictionaryList::~DictionaryList() throw()
  {
  g_log << DEBUG << "Starting DictionaryList destruction" << endlog;

  for (std::list<const Dictionary*>::iterator it = m_dictionaries.begin();
       it != m_dictionaries.end(); ++it)
    delete *it;

  for (StrategyContainer::iterator it = m_strategies.begin(); it != m_strategies.end(); ++it)
    {
    delete it -> second.first;
    delete it -> second.second;
    }

  for (TransformerContainer::iterator it = m_transformers.begin();
       it != m_transformers.end(); ++it)
    delete it -> second;
  }
//------------------------------------------------------------------------------
void DictionaryList::AddDictionary(const Dictionary* ip_dictionary)
  {
  m_dictionaries.push_back(ip_dictionary);
  }
//------------------------------------------------------------------------------
void DictionaryList::AddStrategy(const AbstractStrategy* ip_strategy, const AccessRights* ip_rights)
  {
  assert(m_strategies.find(ip_strategy -> GetStrategyName()) == m_strategies.end());
  m_strategies[ip_strategy -> GetStrategyName()] = StrategyPair(ip_strategy, ip_rights);
  }
//------------------------------------------------------------------------------
void DictionaryList::AddDictionarySet(const UnicodeString& i_name, const UnicodeString& i_descr,
				      const std::list<UnicodeString>& i_items)
  {
  m_dictionary_sets[i_name].first = i_descr;
  m_dictionary_sets[i_name].second.assign(i_items.begin(), i_items.end());
  }
//------------------------------------------------------------------------------
void DictionaryList::AddTransformer(const AbstractTransformer* ip_transformer)
  {
  assert(m_transformers.find(ip_transformer -> GetTransformerName()) == m_transformers.end());
  m_transformers[ip_transformer -> GetTransformerName()] = ip_transformer;
  }
//------------------------------------------------------------------------------
UnicodeString DictionaryList::GetDictionaryListStatistics() const throw()
  {
  return UnicodeString();
  }
//------------------------------------------------------------------------------
DictionaryList::QueryStatus
DictionaryList::FindKeyWords(KeyWordsQueryResult& o_keywords, AbstractController& io_controller,
                             const UnicodeString& i_word, const UnicodeString& i_strategy,
                             const UnicodeString& i_dictionary, const SessionInfo& i_session_info) const throw(DictionaryListException)
  {
  QueryStatus res = OK;

  if (m_strategies.find(i_strategy) == m_strategies.end())
    throw DictionaryListException("UnknownStrategy", std::string("Strategy was not found"));

  try
    {
    io_controller.StartMatch();

    const AbstractStrategy* p_strategy = m_strategies.find(i_strategy) -> second.first;

    std::pair<const Dictionary*, std::list<UnicodeString> > cur_keywords;

    for (std::list<const Dictionary*>::const_iterator it = m_dictionaries.begin();
	 it != m_dictionaries.end(); ++it)
      if (IsDictionaryMatched(**it, i_dictionary, i_session_info))
        {
        (*it) -> FindKeyWords(cur_keywords.second, io_controller, i_word, *p_strategy);

        if (!cur_keywords.second.empty())
          {
	  cur_keywords.first = *it;

	  o_keywords.push_back(cur_keywords);

	  if (i_dictionary == "!")
	    break;
	  }
        if (io_controller.WasMatchInterrupted())
          {
          res = PARTIAL_RESULT;
          break;
          }
	}
    }
  catch(const DictionaryException& ex)
    {
    g_log << ex << endlog;
    throw DictionaryListException("FindKeyWordsFailed", "Failure during keywords searching for some of the dictionaries");
    }

  return res;
  }
//------------------------------------------------------------------------------
DictionaryList::QueryStatus
DictionaryList::FindDefinitions(DefinitionsQueryResult& o_definitions, AbstractController& io_controller,
                                const UnicodeString& i_word, const UnicodeString& i_strategy,
                                const UnicodeString& i_dictionary, const SessionInfo& i_session_info) const throw(DictionaryListException)
  {
  QueryStatus res = OK;

  if (m_strategies.find(i_strategy) == m_strategies.end())
    throw DictionaryListException("UnknownStrategy", std::string("Strategy was not found"));

  try
    {
    io_controller.StartMatch(); // As we will use exact matching to find keywords
    io_controller.StartDefine();

    const AbstractStrategy* p_strategy = m_strategies.find(i_strategy) -> second.first;

    std::pair<const Dictionary*, Dictionary::Definitions> cur_defs;

    for (std::list<const Dictionary*>::const_iterator it = m_dictionaries.begin();
	 it != m_dictionaries.end(); ++it)
      if (IsDictionaryMatched(**it, i_dictionary, i_session_info))
	{
  (*it) -> FindDefinitions(cur_defs.second, io_controller, i_word, *p_strategy, i_session_info.m_mime_types, i_session_info.m_fl_use_mime);

	if (!cur_defs.second.empty())
	  {
	  cur_defs.first = *it;

	  o_definitions.push_back(cur_defs);

	  if (i_dictionary == "!")
	    break;
	  }

        if (io_controller.WasDefineInterrupted())
          {
          res = PARTIAL_RESULT;
          break;
          }
	}
    }
  catch(const DictionaryException& ex)
    {
    g_log << ex << endlog;
    throw DictionaryListException("FindDefinitionsFalied", "Failure during definitions fetching for some of the dictionaries");
    }  

  return res;
  }
//------------------------------------------------------------------------------
bool DictionaryList::IsDictionaryPresent(const UnicodeString& i_name,
					 const SessionInfo& i_session_info) const
  {
  if (i_name == "*" || i_name == "!")
    {
    for (DictionaryContainer::const_iterator it = m_dictionaries.begin();
	 it != m_dictionaries.end(); ++it)
      if (IsDictionaryAllowed(**it, i_session_info))
	return true;

    return false;
    }

  DictionarySetContainer::const_iterator it = m_dictionary_sets.find(i_name);

  if (it != m_dictionary_sets.end())
    return IsDictionarySetAllowed(it -> second, i_session_info);

  for (DictionaryContainer::const_iterator it = m_dictionaries.begin(); it !=
       m_dictionaries.end(); ++it)
    if ((*it) -> GetDictionaryName() == i_name)
      return IsDictionaryAllowed(**it, i_session_info);

  return false;
  }
//------------------------------------------------------------------------------
bool DictionaryList::IsStrategyPresent(const UnicodeString& i_name, const SessionInfo& i_session_info) const
  {
  if (m_strategies.find(i_name) == m_strategies.end())
    return false;

  const AccessRights& rights = *m_strategies.find(i_name) -> second.second;

  return IsAllowedByAccessRights(rights, i_session_info);
  }
//------------------------------------------------------------------------------
bool DictionaryList::IsAllowedByAccessRights(const AccessRights& i_rights, const SessionInfo& i_session_info) const
  {
  bool res = i_rights.m_fl_default_access_policy;  

  for (AccessRights::AccessList::const_iterator it = i_rights.m_access_list.begin();
       it != i_rights.m_access_list.end(); ++it)
    {
    if (it -> m_type == AccessRightsItem::USER_NAME && it -> m_name == i_session_info.m_user ||
        it -> m_type == AccessRightsItem::GROUP_NAME && it -> m_name == i_session_info.m_group)
      {
      res = it -> m_fl_allowed;
      break;
      }
    }

  return res;
  }
//------------------------------------------------------------------------------
bool DictionaryList::IsDictionaryAllowed(const Dictionary& i_dict, const SessionInfo& i_session_info) const
  {
  if (!IsAllowedByAccessRights(i_dict.GetAccessRights(), i_session_info))
    return false;

  if (!i_session_info.m_fl_use_mime)
    return true;

  return i_dict.ProduceOutputMimeType(i_session_info.m_mime_types) != "";
  }
//------------------------------------------------------------------------------
bool DictionaryList::IsDictionarySetAllowed(const DictionarySet& i_dict_set, const SessionInfo& i_session_info) const
  {
  for (std::list<UnicodeString>::const_iterator it = i_dict_set.second.begin(); it != i_dict_set.second.end(); ++it)
    if (IsDictionaryPresent(*it, i_session_info))
      return true;

  return false;
  }
//------------------------------------------------------------------------------
bool DictionaryList::IsDictionaryMatched(const Dictionary& i_dict, const UnicodeString& i_name,
					 const SessionInfo& i_session_info) const
  {
  if (i_name == "*" || i_name == "!" || i_dict.GetDictionaryName() == i_name)
    return IsDictionaryAllowed(i_dict, i_session_info);

  if (m_dictionary_sets.find(i_name) != m_dictionary_sets.end())
    {
    const std::list<UnicodeString>& dict_set = m_dictionary_sets.find(i_name) -> second.second;

    if (std::find(dict_set.begin(), dict_set.end(), i_dict.GetDictionaryName()) != dict_set.end())
      return IsDictionaryAllowed(i_dict, i_session_info);
    }

  return false;
  }
//------------------------------------------------------------------------------
ulong DictionaryList::GetNumberOfAvailableDictionaries(const SessionInfo& i_session_info) const
  {
  ulong num = 0;

  for (DictionarySetContainer::const_iterator it = m_dictionary_sets.begin();
       it != m_dictionary_sets.end(); ++it)
    if (IsDictionarySetAllowed(it -> second, i_session_info))
      ++num;

  for (DictionaryContainer::const_iterator it = m_dictionaries.begin();
       it != m_dictionaries.end(); ++it)
    if (IsDictionaryAllowed(**it, i_session_info))
      ++num;

  return num;
  }
//------------------------------------------------------------------------------
