/*****************************************************************************
 *
 * DictD++, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#ifndef DICTIONARY_LIST_H
#define DICTIONARY_LIST_H

#include "dictionary.h"
//------------------------------------------------------------------------------
class AbstractStrategy;
class AbstractTransformer;
class SessionInfo;
//------------------------------------------------------------------------------
// This is an exception class used to report errors
// for dictionary list.
class DictionaryListException: public GeneralException
  {
  public:
    DictionaryListException(const std::string& i_err_id,
                            const std::string& i_msg):
      GeneralException("DictionaryList", i_err_id, i_msg) {}
  };
//------------------------------------------------------------------------------
// This is an DictionaryList object, that contains all Dictionaries
// and is able to run & collect results on the whole set of contained
// Dictionaries as well as on some subsets.
//
class DictionaryList
  {
  public:
    enum QueryStatus
      {
      FAILED,
      OK,
      PARTIAL_RESULT
      };

    typedef std::list<std::pair<const Dictionary*, Dictionary::Definitions> > DefinitionsQueryResult;
    typedef std::list<std::pair<const Dictionary*, std::list<UnicodeString> > > KeyWordsQueryResult;
    typedef std::pair<UnicodeString, std::list<UnicodeString> > DictionarySet;
    typedef std::list<const Dictionary*> DictionaryContainer;
    typedef std::pair<const AbstractStrategy*, const AccessRights*> StrategyPair;
    typedef std::map<UnicodeString, StrategyPair> StrategyContainer;
    typedef std::map<UnicodeString, DictionarySet> DictionarySetContainer;
    typedef std::map<UnicodeString, const AbstractTransformer*> TransformerContainer;

    DictionaryList() {}

    ~DictionaryList() throw();

    void AddDictionary(const Dictionary* ip_dictionary);

    void AddDictionarySet(const UnicodeString& i_name, const UnicodeString& i_descr,
                          const std::list<UnicodeString>& i_items);

    void AddStrategy(const AbstractStrategy* ip_strategy, const AccessRights* ip_access_rights);

    void AddTransformer(const AbstractTransformer* ip_transformer);

    UnicodeString GetDictionaryListStatistics() const throw();

    QueryStatus FindKeyWords(KeyWordsQueryResult& o_keywords, AbstractController& io_controller,
                             const UnicodeString& i_word, const UnicodeString& i_strategy_name,
                             const UnicodeString& i_dictionaries_set,
                             const SessionInfo& i_session_info) const throw(DictionaryListException);

    QueryStatus FindDefinitions(DefinitionsQueryResult& o_definitions, AbstractController& io_controller,
                                const UnicodeString& i_word, const UnicodeString& i_strategy_name,
                                const UnicodeString& i_dictionaries_set,
                                const SessionInfo& i_session_info) const throw(DictionaryListException);

    //
    // Utility functions to access dictionaries/dictionary sets/strategies:
    //

    //
    // Direct read-only access to structures.
    //
    const DictionaryContainer& GetDictionaryContainer() const { return m_dictionaries; }

    const StrategyContainer& GetStrategyContainer() const { return m_strategies; }

    const DictionarySetContainer& GetDictionarySetContainer() const { return m_dictionary_sets; }

    const TransformerContainer& GetTransformerContainer() const { return m_transformers; }

    /// Returns true if there's a strategy named `i_name' and 
    /// there's access to it with `i_session info'
    bool IsStrategyPresent(const UnicodeString& i_name, const SessionInfo& i_session_info) const;

    /// Returns true if there's a dictionary named `i_name' and 
    /// there's access to it with `i_session info'
    bool IsDictionaryPresent(const UnicodeString& i_name, const SessionInfo& i_session_info) const;

    /// Returns true if `i_dict' is allowed to access with `i_session_info'
    bool IsDictionaryAllowed(const Dictionary& i_dict, const SessionInfo& i_session) const;

    /// Returns true if `i_dict_set' is allowed to access with `i_session_info'
    bool IsDictionarySetAllowed(const DictionarySet& i_dict_set, const SessionInfo& i_session) const;

    /// Returns true if `i_dict' corresponds to `i_name' and `i_session_info' allows access
    bool IsDictionaryMatched(const Dictionary& i_dict, const UnicodeString& i_name, const SessionInfo& i_session_info) const;

    /// Returns number of available dictionaries and dictionary sets
    ulong GetNumberOfAvailableDictionaries(const SessionInfo& i_session_info) const;

  private:
    DictionaryContainer m_dictionaries;
    StrategyContainer m_strategies;
    DictionarySetContainer m_dictionary_sets;
    TransformerContainer m_transformers;

    bool IsAllowedByAccessRights(const AccessRights& i_rights, const SessionInfo& i_session_info) const;
  };
//------------------------------------------------------------------------------

#endif // DICTIONARY_LIST_H
