/*****************************************************************************
 *
 * DictD++, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#include "stdafx.h"
#include "server.h"
#include "server_config.h"
#include "server_socket.h"
#include "security_manager.h"
#include "control_thread.h"

#include <dictionary_list.h>

#include <exact_match_strategy.h>
#include <prefix_match_strategy.h>
#include <access_rights.h>

#include <xercesc/sax2/XMLReaderFactory.hpp>

//--------------------------------------------------------------------
// ServerXMLErrorHandler implementation: just throws the exception further
//--------------------------------------------------------------------
void ServerXMLErrorHandler::error(const xercesc::SAXParseException& ex)
  {
  throw ex;
  }
//--------------------------------------------------------------------
// Server implementation
//--------------------------------------------------------------------
Server::Server()
  {
  mp_security_manager = new SecurityManager(*this);
  m_port = 2628;
  m_max_threads = 32;
  m_time_accuracy = 5;

  // These strategies are required by RFC2229; all other are optional
  // and are specified in config in order to allow access rights specification
  m_dictionary_list.AddStrategy(new ExactMatchStrategy(), new AccessRights);
  m_dictionary_list.AddStrategy(new PrefixMatchStrategy(), new AccessRights);

  mp_parser = NULL;
  }
//--------------------------------------------------------------------
Server::~Server()
  {
  delete mp_security_manager;
  }
//--------------------------------------------------------------------
void Server::ParseConfigAndInit(const std::string& i_file_name) throw(ServerException)
  {
  g_log << INFO << "Starting config parsing" << endlog;

  try
    {
    mp_parser = xercesc::XMLReaderFactory::createXMLReader();

    mp_parser -> setFeature(xercesc::XMLUni::fgSAX2CoreValidation, true);
    mp_parser -> setFeature(xercesc::XMLUni::fgXercesDynamic, false);
    mp_parser -> setFeature(xercesc::XMLUni::fgSAX2CoreNameSpaces, true);
    mp_parser -> setFeature(xercesc::XMLUni::fgXercesSchema, true);
    mp_parser -> setFeature(xercesc::XMLUni::fgXercesSchemaFullChecking, true);
    mp_parser -> setFeature(xercesc::XMLUni::fgSAX2CoreNameSpacePrefixes, true);

    XMLCh* p_schema_file = xercesc::XMLString::transcode("dictd++.xsd");

    mp_parser -> setProperty(xercesc::XMLUni::fgXercesSchemaExternalNoNameSpaceSchemaLocation, p_schema_file);

    ServerXMLContentHandler content_handler(this);
    ServerXMLErrorHandler error_handler;

    mp_parser -> setErrorHandler(&error_handler);

    // First pass is to check config only
    mp_parser -> parse(i_file_name.c_str());

    // Second pass is to perform initialization
    m_handlers.push_back(&content_handler);
    mp_parser -> setContentHandler(&content_handler);
    mp_parser -> parse(i_file_name.c_str());

    content_handler.ClearHandlersToRemove();

    //g_log << INFO << "Config parsing has been finished, " << mp_parser -> getErrorCount() << " error(s)" << endlog;
    g_log << INFO << "Config parsing has been finished" << endlog;

    xercesc::XMLString::release(&p_schema_file);

    delete mp_parser;
    mp_parser = NULL;
    }
  catch (const xercesc::SAXParseException& ex)
    {
    throw ServerException("WrongConfig", std::string("Parsing config file \"")
					 + AsString(ex.getSystemId()) + "\" failed at line "
					 + boost::lexical_cast<std::string>(ex.getLineNumber())
					 + ", column "
					 + boost::lexical_cast<std::string>(ex.getColumnNumber())
					 + ": " + AsString(ex.getMessage())
					 + " Please make sure config is correct and fully conforms to the schema file.");
    }
  catch (const xercesc::XMLException& ex)
    {
    throw ServerException("WrongConfig", std::string("Parsing config file \"")
					 + ex.getSrcFile() + "\" failed at line "
					 + boost::lexical_cast<std::string>(ex.getSrcLine())
					 + ": " + AsString(ex.getMessage())
					 + " Please make sure config is correct and fully conforms to the schema file.");
    }

  try
    {
    GetSecurityManager().Validate();
    }
  catch (const SecurityManagerException& ex)
    {
    g_log << CRIT << ex << endlog;
    throw ServerException("SecurityInfoValidationFailed", "Security dependencies validation failed, please check your config");
    }

  try
    {
    const DictionaryList::DictionaryContainer& cont = GetDictionaryList().GetDictionaryContainer();

    for (DictionaryList::DictionaryContainer::const_iterator it = cont.begin(); it != cont.end(); ++it)
      (*it) -> Validate();
    }
  catch (const DictionaryException& ex)
    {
    g_log << CRIT << ex << endlog;
    throw ServerException("DictionaryValidationFailed", "Dictionaries dependencies validation failed, please check your config");
    }
  }
//--------------------------------------------------------------------
void Server::Run() throw()
  {
  mp_control_thread = new ControlThread(m_time_accuracy);
  boost::thread* p_ctrl_thread = new boost::thread(ControlThreadWrapper(mp_control_thread));

  g_log << INFO << "Ready to serve requests" << endlog;

  ServerSocket socket(m_port);

  while (true)
    {
    int res;

    while (!(res = socket.Select(Socket::SELECT_READ | Socket::SELECT_ERROR, m_time_accuracy)))
      if (mp_control_thread -> IsShuttingDown())
        break;

    if (mp_control_thread -> IsShuttingDown())
      {
      g_log << INFO << "Shutdown has been requested, connections are not accepted anymore" << endlog;
      socket.Close();
      break;
      }

    if (res == SOCKET_ERROR || (res & Socket::SELECT_ERROR))
      {
      g_log << CRIT << "Server socket error, terminating server" << endlog;
      mp_control_thread -> SetIsShuttingDown();
      socket.Close();
      break;
      }

    if (!(res & Socket::SELECT_READ))
      continue;

    Socket new_socket;

    try
      {
      socket.Accept(new_socket);
      }
    catch(const SocketException& ex)
      {
      g_log << ERR << ex << endlog;
      g_log << ERR << "Was not able to accept connection" << endlog;
      continue;
      }

    SessionInfo info;

    try
      {
      if (!GetSecurityManager().IsConnectionAllowed(new_socket, info.m_user))
        {
        g_log << TRACE << "Connection is refused" << endlog;
        new_socket.SetNonBlocking(true);
        new_socket << "530 Access denied\r\n";
        new_socket.FlushOutput();
        new_socket.Close();
        continue;
        }
      }
    catch(const SocketException& ex)
      {
      g_log << TRACE << ex << endlog;
      g_log << TRACE << "Was not able to send refusal explanation to the client" << endlog;
      continue;
      }

    boost::mutex::scoped_lock lock(m_mutex);

    if (m_threads.size() >= m_max_threads)
      {
      g_log << WARN << "Cannot accept new connection, number of running threads is already " << (ulong)m_threads.size() << endlog;
      try
        {
        new_socket.SetNonBlocking(true);
        new_socket << "420 Server temporarily unavailable, number of running threads is already " << boost::lexical_cast<std::string>((ulong)m_threads.size()) << "\r\n";
        new_socket.FlushOutput();
        new_socket.Close();
        }
      catch(const SocketException& ex)
        {
        g_log << WARN << ex << endlog;
        g_log << WARN << "Was not able to send refusal explanation to the client" << endlog;
        }
      continue;
      }

    ServerThread* p_server_thread = new ServerThread(new_socket, *this);
    m_threads[p_server_thread] = new boost::thread(ServerThreadWrapper(p_server_thread));
    }

  g_log << INFO << "Waiting for all sessions to finish, timeout is " << GetTimeAccuracy() << " seconds" << endlog;
  do
    {
    boost::mutex::scoped_lock lock(m_mutex);
    if (m_threads.empty())
      break;
    boost::thread().yield();
    }
  while(true);

  p_ctrl_thread -> join();
  delete mp_control_thread;
  delete p_ctrl_thread;

  g_log << INFO << "All threads were finished, exiting" << endlog;
  }
//--------------------------------------------------------------------
UnicodeString Server::GetStatistics() const throw()
  {
  return m_dictionary_list.GetDictionaryListStatistics();
  }
//--------------------------------------------------------------------
void Server::SetIsShuttingDown() throw()
  {
  mp_control_thread -> SetIsShuttingDown();
  }
//--------------------------------------------------------------------

