/*****************************************************************************
 *
 * DictD++, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#include "stdafx.h"

#include "server_config.h"
#include "server.h"
#include "security_manager.h"

#include <dictionary.h>
#include <dictionary_list.h>
#include <memory_mapped_file.h>
#include <dictd_index.h>
#include <plain_storage.h>
#include <dzip_storage.h>
#include <xml_transformer.h>
#include <convertion.h>
#include <access_rights.h>

#include <wrapper_strategy.h>
#include <regex_match_strategy.h>
#include <substring_match_strategy.h>
#include <suffix_match_strategy.h>
#include <levinshtein_match_strategy.h>
#include <top_levinshtein_match_strategy.h>

#include <xercesc/sax2/SAX2XMLReader.hpp>

#include <sys/stat.h>
#include <time.h>

//--------------------------------------------------------------------
// ConfigHandlerBase implementation:
//--------------------------------------------------------------------
void ConfigHandlerBase::startElement(const XMLCh* const uri, const XMLCh* const localname, const XMLCh* const qname,
				     xercesc::Attributes& attributes)
  {
  m_text.remove();
  }
//--------------------------------------------------------------------
void ConfigHandlerBase::characters(const XMLCh* const chars, const unsigned int length)
  {
  m_text.append(chars, length);
  }
//--------------------------------------------------------------------
void ConfigHandlerBase::endElement(const XMLCh* const uri, const XMLCh* const localname, const XMLCh* const qname)
  {
  m_text.remove();
  PopHandler();
  }
//--------------------------------------------------------------------
void ConfigHandlerBase::ClearHandlersToRemove()
  {
  for (std::list<ConfigHandlerBase*>::iterator it = mp_server -> m_handlers2remove.begin();
       it != mp_server -> m_handlers2remove.end(); ++it)
    delete *it;

  mp_server -> m_handlers2remove.clear();
  }
//--------------------------------------------------------------------
void ConfigHandlerBase::SetNewHandler(ConfigHandlerBase* ip_new_handler)
  {
  ClearHandlersToRemove();
  mp_server -> m_handlers.push_back(ip_new_handler);
  mp_server -> mp_parser -> setContentHandler(ip_new_handler);
  }
//--------------------------------------------------------------------
void ConfigHandlerBase::PopHandler()
  {
  assert(mp_server -> m_handlers.size());

  mp_server -> m_handlers2remove.push_back(mp_server -> m_handlers.back());
  mp_server -> m_handlers.pop_back();
  mp_server -> mp_parser -> setContentHandler(mp_server -> m_handlers.back());
  }
//--------------------------------------------------------------------
ConfigHandlerBase* ConfigHandlerBase::GetParent()
  {
  std::list<ConfigHandlerBase*>::iterator it = std::find(mp_server -> m_handlers.begin(),
							 mp_server -> m_handlers.end(), this);
  assert(it != mp_server -> m_handlers.begin() && it != mp_server -> m_handlers.end());

  return *--it;
  }
//--------------------------------------------------------------------
DictionaryList& ConfigHandlerBase::GetDictionaryList()
  {
  return mp_server -> m_dictionary_list;
  }
//--------------------------------------------------------------------
void ConfigHandlerBase::CopyAttributes(const xercesc::Attributes& i_attributes)
  {
  m_attributes.clear();

  for (ulong i = 0; i < i_attributes.getLength(); ++i)
    m_attributes[AsString(i_attributes.getLocalName(i))] = i_attributes.getValue(i);
  }
//--------------------------------------------------------------------
// ServerXMLContentHandler implementation:
//--------------------------------------------------------------------
void ServerXMLContentHandler::startElement(const XMLCh* const uri, const XMLCh* const localname,
					   const XMLCh* const qname, const xercesc::Attributes& attributes)
  {
  m_text.remove();
  assert(AsString(localname) == "Config");

  SetNewHandler(new ConfigHandler(mp_server));
  }
//--------------------------------------------------------------------
void ServerXMLContentHandler::endElement(const XMLCh* const uri, const XMLCh* const localname,
					 const XMLCh* const qname)
  {
  assert(AsString(localname) == "Config");
  }
//--------------------------------------------------------------------
// Config handler implementation
//--------------------------------------------------------------------
void ConfigHandler::startElement(const XMLCh* const uri, const XMLCh* const localname,
				 const XMLCh* const qname, const xercesc::Attributes& attributes)
  {
  m_text.remove();
  std::string element = AsString(localname);

  if (element == "Server")
    SetNewHandler(new ServerHandler(mp_server));
  else if (element == "Groups")
    SetNewHandler(new GroupsHandler(mp_server));
  else if (element == "Users")
    SetNewHandler(new UsersHandler(mp_server));
  else if (element == "TransformerList")
    SetNewHandler(new TransformerListHandler(mp_server));
  else if (element == "StrategyList")
    SetNewHandler(new StrategyListHandler(mp_server));
  else if (element == "DictionaryList")
    SetNewHandler(new DictionaryListHandler(mp_server));
  else
    assert(0);
  }
//--------------------------------------------------------------------
// ServerHandler implementation
//--------------------------------------------------------------------
void ServerHandler::endElement(const XMLCh* const uri, const XMLCh* const localname, const XMLCh* const qname)
  {
  m_text.trim();
  std::string element = AsString(localname);

  if (element == "MaxThreads")
    mp_server -> m_max_threads = boost::lexical_cast<unsigned>(AsString(m_text));
  else if (element == "Port")
    mp_server -> m_port = boost::lexical_cast<unsigned>(AsString(m_text));
  else if (element == "TimeAccuracy")
    mp_server -> m_time_accuracy = boost::lexical_cast<unsigned>(AsString(m_text));
  else if (element == "CachePath")
    mp_server -> m_cache_path = AsString(m_text);
  else if (element == "ServerInfo")
    mp_server -> m_server_info = m_text;
  else if (element == "Server")
    return PopHandler();
  else
    assert(0);

  m_text.remove();
  }
//--------------------------------------------------------------------
// TransformerListHandler implementation:
//--------------------------------------------------------------------
void TransformerListHandler::startElement(const XMLCh* const uri, const XMLCh* const localname,
					  const XMLCh* const qname, const xercesc::Attributes& attributes)
  {
  m_text.remove();
  std::string element = AsString(localname);

  if (element == "Transformer")
    SetNewHandler(new TransformerHandler(mp_server));
  else
    assert(0);
  }
//--------------------------------------------------------------------
// TransformerHandler implementation:
//--------------------------------------------------------------------
TransformerHandler::TransformerHandler(Server* ip_server): ConfigHandlerBase(ip_server), mp_transformer(NULL) {}
//--------------------------------------------------------------------
void TransformerHandler::startElement(const XMLCh* const uri, const XMLCh* const localname,
				      const XMLCh* const qname, const xercesc::Attributes& attributes)
  {
  m_text.remove();
  std::string element = AsString(localname);

  if (element == "Transform")
    SetNewHandler(new TransformHandler(mp_server));
  else
    assert (element == "Name");
  }
//--------------------------------------------------------------------
void TransformerHandler::endElement(const XMLCh* const uri, const XMLCh* const localname, const XMLCh* const qname)
  {
  m_text.trim();
  std::string element = AsString(localname);

  if (element == "Name")
    {
#ifdef DICTDPP_USE_XALAN
    if (m_text == "XML")
      mp_transformer = new XMLTransformer();
    else
#endif
      throw ServerException("UnknownTransformer", "Unknown transformer name");
    }
  else if (element == "Transformer")
    {
    GetDictionaryList().AddTransformer(mp_transformer);
    return PopHandler();
    }

  m_text.remove();
  }
//--------------------------------------------------------------------
// TransformHandler implementation:
//--------------------------------------------------------------------
void TransformHandler::endElement(const XMLCh* const uri, const XMLCh* const localname, const XMLCh* const qname)
  {
  m_text.trim();
  std::string element = AsString(localname);

  if (element == "Name")
    m_name = m_text;
  else if (element == "InputMimeType")
    m_in_mime_type = AsString(m_text);
  else if (element == "OutputMimeType")
    m_out_mime_type = AsString(m_text);
  else if (element == "TransformParams")
    m_params = AsString(m_text);
  else if (element == "Transform")
    {
    dynamic_cast<TransformerHandler&>(*GetParent()).mp_transformer ->
      AddTransform(m_name, m_in_mime_type, m_out_mime_type, m_params);
    return PopHandler();
    }
  else
    assert(0);

  m_text.remove();
  }
//--------------------------------------------------------------------
// StrategyListHandler implementation:
//--------------------------------------------------------------------
void StrategyListHandler::startElement(const XMLCh* const uri, const XMLCh* const localname,
				       const XMLCh* const qname, const xercesc::Attributes& attributes)
  {
  m_text.remove();
  std::string element = AsString(localname);

  if (element == "Strategy")
    SetNewHandler(new StrategyHandler(mp_server));
  else
    assert(0);
  }
//--------------------------------------------------------------------
// StrategyHandler implementation:
//--------------------------------------------------------------------
StrategyHandler::StrategyHandler(Server* ip_server):
  ConfigHandlerBase(ip_server), mp_access_rights(new AccessRights) {}
//--------------------------------------------------------------------
void StrategyHandler::startElement(const XMLCh* const uri, const XMLCh* const localname,
				   const XMLCh* const qname, const xercesc::Attributes& attributes)
  {
  m_text.remove();
  std::string element = AsString(localname);

  if (element == "StrategyAlgorithm")
    CopyAttributes(attributes);
  else if (element == "Access")
    {
    CopyAttributes(attributes);

    if (m_attributes.find("default-policy") == m_attributes.end())
      throw ServerException("WrongConfig", std::string("default-policy was not specified"));

    mp_access_rights -> m_fl_default_access_policy = (m_attributes.find("default-policy") -> second == "allow");
    SetNewHandler(new AccessHandler(mp_server, mp_access_rights));
    }
  else
    assert (element == "Name" || element == "Description");
  }
//--------------------------------------------------------------------
void StrategyHandler::endElement(const XMLCh* const uri, const XMLCh* const localname, const XMLCh* const qname)
  {
  m_text.trim();
  std::string element = AsString(localname);

  if (element == "Name")
    m_name = m_text;
  else if (element == "Description")
    m_descr = m_text;
  else if (element == "StrategyAlgorithm")
    {
    if (m_text == "Levinshtein")
      {
      if (m_attributes.find("distance") == m_attributes.end())
	      throw ServerException("WrongStrategyParams", "Levinshtein \'distance\' parameter was not found");

      mp_strategy = new LevinshteinMatchStrategy(boost::lexical_cast<ulong>(AsString(m_attributes["distance"])));
      }
    else if (m_text == "Top-Levinshtein")
      {
      if (m_attributes.find("top") == m_attributes.end())
        throw ServerException("WrongStrategyParams", "Top-Levinshtein \'top\' parameter was not found");

      if (m_attributes.find("strict") == m_attributes.end())
        throw ServerException("WrongStrategyParams", "Top-Levinshtein \'strict\' parameter was not found");

      if (m_attributes.find("max_distance") == m_attributes.end())
        throw ServerException("WrongStrategyParams", "Top-Levinshtein \'max_distance\' parameter was not found");

      mp_strategy = new TopLevinshteinMatchStrategy(boost::lexical_cast<ulong>(AsString(m_attributes["top"])),
      m_attributes["strict"].toLower() == "true", boost::lexical_cast<ulong>(AsString(m_attributes["max_distance"])));
      }
    else if (m_text == "regex")
      mp_strategy = new RegexMatchStrategy();
    else if (m_text == "substring")
      mp_strategy = new SubstringMatchStrategy();
    else if (m_text == "suffix")
      mp_strategy = new SuffixMatchStrategy();
    else
      throw ServerException("UnknownStrategy", "Unknown strategy algorithm name");
    }
  else if (element == "Strategy")
    {
    GetDictionaryList().AddStrategy(new WrapperStrategy(m_name, m_descr, mp_strategy), mp_access_rights);
    return PopHandler();
    }

  m_text.remove();
  }
//--------------------------------------------------------------------
// DictionaryListHandler implementation:
//--------------------------------------------------------------------
void DictionaryListHandler::startElement(const XMLCh* const uri, const XMLCh* const localname,
					 const XMLCh* const qname, const xercesc::Attributes& attributes)
  {
  m_text.remove();
  std::string element = AsString(localname);

  if (element == "Dictionary")
    SetNewHandler(new DictionaryHandler(mp_server));
  else if (element == "DictionarySet")
    SetNewHandler(new DictionarySetHandler(mp_server));
  else
    assert(0);
  }
//--------------------------------------------------------------------
// DictionaryHandler implementation:
//--------------------------------------------------------------------
DictionaryHandler::DictionaryHandler(Server* ip_server): ConfigHandlerBase(ip_server),
  mp_convertions_list(new std::list<Convertion*>), mp_access_rights(new AccessRights) {}
//--------------------------------------------------------------------
void DictionaryHandler::startElement(const XMLCh* const uri, const XMLCh* const localname,
				     const XMLCh* const qname, const xercesc::Attributes& attributes)
  {
  m_text.remove();
  std::string element = AsString(localname);

  if (element == "AllowedConvertions")
    SetNewHandler(new AllowedConvertionsHandler(mp_server));
  else if (element == "Index" || element == "Body")
    CopyAttributes(attributes);
  else if (element == "Access")
    {
    CopyAttributes(attributes);

    if (m_attributes.find("default-policy") == m_attributes.end())
      throw ServerException("WrongConfig", std::string("default-policy was not specified"));

    mp_access_rights -> m_fl_default_access_policy = (m_attributes.find("default-policy") -> second == "allow");
    SetNewHandler(new AccessHandler(mp_server, mp_access_rights));
    }
  }
//--------------------------------------------------------------------
void DictionaryHandler::endElement(const XMLCh* const uri, const XMLCh* const localname,
				   const XMLCh* const qname)
  {
  m_text.trim();
  std::string element = AsString(localname);

  if (element == "Name")
    m_dict_name = m_text;
  else if (element == "Description")
    m_dict_descr = m_text;
  else if (element == "DefaultMimeType")
    m_dict_def_mime_type = AsString(m_text);
  else if (element == "Index")
    HandleIndex(AsString(m_text));
  else if (element == "Body")
    HandleBody(AsString(m_text));
  else if (element == "Dictionary")
    {
    Dictionary* p_dict = new Dictionary(m_dict_name, m_dict_descr, m_dict_def_mime_type,
					mp_index, mp_storage, mp_convertions_list, mp_access_rights);
    p_dict -> InitConvertions();
    GetDictionaryList().AddDictionary(p_dict);
    return PopHandler();
    }
  else
    assert(0);

  m_text.remove();
  }
//--------------------------------------------------------------------
void DictionaryHandler::HandleIndex(const std::string& i_name)
  {
  if (m_attributes.find("type") == m_attributes.end())
    throw ServerException("WrongConfig", std::string("Index type was not specified"));

  if (m_attributes["type"] != "DictD")
    throw ServerException("WrongConfig", std::string("Index type other than \"DictD\" is not supported"));

  if (m_attributes.find("encoding") == m_attributes.end())
    throw ServerException("WrongConfig", std::string("Index encoding was not specified"));

  if (m_attributes.find("locale") == m_attributes.end())
    throw ServerException("WrongConfig", std::string("Locale was not specified"));

  bool fl_extended = false;

  if (m_attributes.find("cache_type") != m_attributes.end())
    fl_extended = (m_attributes["cache_type"] == "extended");

  // OK, now we can try to construct smth ...
  struct stat stat_info;
  std::string time_str, id;

  if (stat(i_name.c_str(), &stat_info))
    throw ServerException("WrongConfig", std::string("Cannot \"stat\" file ") + i_name);

  time_str = ctime(&stat_info.st_mtime);
  time_str.erase(--time_str.end());
  id = std::string("File size: ") + boost::lexical_cast<std::string>((ulong)stat_info.st_size) + ", timestamp: " + time_str;

  std::string cache_name = ProduceCacheName(i_name);

  // We should create file if it doesn't exist
  if (stat(cache_name.c_str(), &stat_info))
    std::fstream(cache_name.c_str(), std::ios::out);

  std::fstream cache_file(cache_name.c_str(), std::ios::in | std::ios::out | std::ios::binary);

  DictdIndex* p_index = NULL;

  try
    {
    MemoryMappedFile* p_map_index_file = new MemoryMappedFile(i_name);

    DictdIndexCacher cacher(AsString(m_attributes["locale"]), AsString(m_attributes["encoding"]), *p_map_index_file, id, fl_extended);

    if (!cacher.IsCacheValid(cache_file))
      {
      cache_file.close();
      remove(cache_name.c_str());
      std::fstream(cache_name.c_str(), std::ios::out);
      cache_file.open(cache_name.c_str(), std::ios::in | std::ios::out | std::ios::binary);
      cacher.CreateCache(cache_file);
      }

    cache_file.close();

    MemoryMappedFile* p_map_cache_file = new MemoryMappedFile(cache_name);

    p_index = new DictdIndex(p_map_index_file, p_map_cache_file);
    }
  catch(const IndexException& ex)
    {
    g_log << ERR << "Exception during index construction for index " << i_name << endlog;
    g_log << ERR << ex << endlog;
    g_log << ERR << "Please check that index is correct and parameters in config file correspond to this index" << endlog;
    g_log << ERR << "Use dictsort++ to recreate index with appropriate encoding and locale if needed" << endlog;

    cache_file.close();
    remove(cache_name.c_str()); // File may be partially constructed or incorrect
    throw;
    }

  mp_index = p_index;
  }
//--------------------------------------------------------------------
void DictionaryHandler::HandleBody(const std::string& i_name)
  {
  if (m_attributes.find("type") == m_attributes.end())
    throw ServerException("WrongConfig", std::string("Body type was not specified"));

  if (m_attributes["type"] != "dictzip" &&
      m_attributes["type"] != "plain")
      throw ServerException("WrongConfig", std::string("Unknown body type: \"") +
	      AsString(m_attributes["type"]) + "\". Body types other than \"dictzip\" and \"plain\" are not supported");

  if (m_attributes.find("encoding") == m_attributes.end())
    throw ServerException("WrongConfig", std::string("Body encoding was not specified"));

  if (m_attributes.find("mime-type") == m_attributes.end())
    throw ServerException("WrongConfig", std::string("MIME type was not specified"));

  // OK, now we can try to construct smth ...
  struct stat stat_info;

  if (stat(i_name.c_str(), &stat_info))
    throw ServerException("WrongConfig", std::string("Cannot \"stat\" file") + i_name);

  if (m_attributes["type"] == "plain")
    mp_storage = new PlainStorage(AsString(m_attributes["encoding"]), AsString(m_attributes["mime-type"]), new MemoryMappedFile(i_name));
  else if (m_attributes["type"] == "dictzip")
    mp_storage = new DZipStorage(AsString(m_attributes["encoding"]), AsString(m_attributes["mime-type"]), new MemoryMappedFile(i_name));
  }
//--------------------------------------------------------------------
std::string DictionaryHandler::ProduceCacheName(const std::string& i_name) const
  {
  // we should substitute the path to the index file by the m_cache_path
  // ads
  std::string::size_type end = i_name.length() - 1;
  for (; end > 0 && i_name[end] != '\\' && i_name[end] != '/'; --end);

  std::string cache_path = mp_server -> GetCachePath();
#ifdef _WIN32
  if (*--cache_path.end() != '\\')
    cache_path += '\\';
#else
  if (*--cache_path.end() != '/')
    cache_path += '/';
#endif

  if (!end)
    return cache_path + i_name + ".cache";

  return cache_path + i_name.substr(end + 1) + ".cache";
  }
//--------------------------------------------------------------------
// AllowedConvertionsHandler implementation:
//--------------------------------------------------------------------
void AllowedConvertionsHandler::startElement(const XMLCh* const uri, const XMLCh* const localname,
					     const XMLCh* const qname, const xercesc::Attributes& attributes)
  {
  m_text.remove();
  std::string element = AsString(localname);

  if (element == "Convertion")
    SetNewHandler(new ConvertionHandler(mp_server));
  else
    assert(0);
  }
//--------------------------------------------------------------------
// ConvertionHandler implementation:
//--------------------------------------------------------------------
ConvertionHandler::ConvertionHandler(Server* ip_server): ConfigHandlerBase(ip_server)
  {
  mp_convertion = new Convertion(GetDictionaryList().GetTransformerContainer());
  }
//--------------------------------------------------------------------
void ConvertionHandler::startElement(const XMLCh* const uri, const XMLCh* const localname,
				     const XMLCh* const qname, const xercesc::Attributes& attributes)
  {
  m_text.remove();
  std::string element = AsString(localname);

  if (element == "ApplyTransform")
    CopyAttributes(attributes);
  else
    assert(0);
  }
//--------------------------------------------------------------------
void ConvertionHandler::endElement(const XMLCh* const uri, const XMLCh* const localname,
				   const XMLCh* const qname)
  {
  std::string element = AsString(localname);

  if (element == "ApplyTransform")
    mp_convertion -> AddTransformStep(UnicodeString(m_attributes["transformer"]),
				      UnicodeString(m_attributes["transform"]));
  else if (element == "Convertion")
    {
    dynamic_cast<DictionaryHandler&>(*GetParent() -> GetParent()).mp_convertions_list -> push_back(mp_convertion);
    return PopHandler();
    }
  else
    assert(0);

  m_text.remove();
  }
//--------------------------------------------------------------------
// DictionarySetHandler implementation:
//--------------------------------------------------------------------
void DictionarySetHandler::endElement(const XMLCh* const uri, const XMLCh* const localname,
				      const XMLCh* const qname)
  {
  m_text.trim();
  std::string element = AsString(localname);

  if (element == "Name")
    m_name = m_text;
  else if (element == "Description")
    m_descr = m_text;
  else if (element == "Item")
    m_items.push_back(m_text);
  else if (element == "DictionarySet")
    {
    GetDictionaryList().AddDictionarySet(m_name, m_descr, m_items);
    return PopHandler();
    }
  else
    assert(0);

  m_text.remove();
  }
//--------------------------------------------------------------------
// GroupsHandler implementation:
//--------------------------------------------------------------------
void GroupsHandler::startElement(const XMLCh* const uri, const XMLCh* const localname,
                                 const XMLCh* const qname, const xercesc::Attributes& attributes)
  {
  m_text.remove();
  std::string element = AsString(localname);

  if (element == "Group")
    {
    CopyAttributes(attributes);

    if (m_attributes.find("name") == m_attributes.end())
      throw ServerException("WrongConfig", std::string("name was not specified"));

    SetNewHandler(new GroupHandler(mp_server, &mp_server -> GetSecurityManager().m_groups[AsString(m_attributes["name"])]));
    }
  else
    assert(0);
  }
//--------------------------------------------------------------------
void GroupsHandler::endElement(const XMLCh* const uri, const XMLCh* const localname,
                               const XMLCh* const qname)
  {
  std::string element = AsString(localname);

  if (element == "Groups")
    return PopHandler();
  else
    assert(0);

  m_text.remove();
  }
//--------------------------------------------------------------------
// UsersHandler implementation:
//--------------------------------------------------------------------
void UsersHandler::startElement(const XMLCh* const uri, const XMLCh* const localname,
                                const XMLCh* const qname, const xercesc::Attributes& attributes)
  {
  m_text.remove();
  std::string element = AsString(localname);

  if (element == "User")
    {
    CopyAttributes(attributes);

    if (m_attributes.find("name") == m_attributes.end())
      throw ServerException("WrongConfig", std::string("name was not specified"));

    if (m_attributes.find("password") == m_attributes.end())
      throw ServerException("WrongConfig", std::string("password was not specified"));

    if (m_attributes.find("group") == m_attributes.end())
      throw ServerException("WrongConfig", std::string("group was not specified"));

    UserInfo* p_user_info = &mp_server -> GetSecurityManager().m_users[AsString(m_attributes["name"])];

    p_user_info -> m_group = AsString(m_attributes["group"]);
    p_user_info -> m_password = AsString(m_attributes["password"]);
    // Will override access info later if necessary
    p_user_info -> m_access_info = mp_server -> GetSecurityManager().m_groups[p_user_info -> m_group];

    SetNewHandler(new UserHandler(mp_server, p_user_info));
    }
  else
    assert(0);
  }
//--------------------------------------------------------------------
void UsersHandler::endElement(const XMLCh* const uri, const XMLCh* const localname,
                              const XMLCh* const qname)
  {
  std::string element = AsString(localname);

  if (element == "Users")
    return PopHandler();
  else
    assert(0);

  m_text.remove();
  }
//--------------------------------------------------------------------
// GroupHandler implementation:
//--------------------------------------------------------------------
GroupHandler::GroupHandler(Server* ip_server, AccessInfo* iop_access_info):
  ConfigHandlerBase(ip_server), mp_access_info(iop_access_info) {}
//--------------------------------------------------------------------
void GroupHandler::startElement(const XMLCh* const uri, const XMLCh* const localname,
                                const XMLCh* const qname, const xercesc::Attributes& attributes)
  {
  m_text.remove();
  std::string element = AsString(localname);

  if (element == "NetAccess")
    {
    CopyAttributes(attributes);

    if (m_attributes.find("default-policy") == m_attributes.end())
      throw ServerException("WrongConfig", std::string("default-policy was not specified"));

    mp_access_info -> m_fl_default_access_policy = (m_attributes.find("default-policy") -> second == "allow");
    
    SetNewHandler(new NetAccessHandler(mp_server, mp_access_info));
    }
  else if (element == "QueryOptions")
    SetNewHandler(new QueryOptionsHandler(mp_server, mp_access_info));
  else
    assert(0);
  }
//--------------------------------------------------------------------
void GroupHandler::endElement(const XMLCh* const uri, const XMLCh* const localname,
                              const XMLCh* const qname)
  {
  std::string element = AsString(localname);

  if (element == "Group")
    return PopHandler();
  else
    assert(element == "NetAccess" || element == "QueryOptions");

  m_text.remove();
  }
//--------------------------------------------------------------------
// UserHandler implementation:
//--------------------------------------------------------------------
UserHandler::UserHandler(Server* ip_server, UserInfo* iop_user_info):
  ConfigHandlerBase(ip_server), mp_user_info(iop_user_info) {}
//--------------------------------------------------------------------
void UserHandler::startElement(const XMLCh* const uri, const XMLCh* const localname,
                               const XMLCh* const qname, const xercesc::Attributes& attributes)
  {
  m_text.remove();
  std::string element = AsString(localname);

  if (element == "NetAccess")
    {
    CopyAttributes(attributes);

    if (m_attributes.find("default-policy") == m_attributes.end())
      throw ServerException("WrongConfig", std::string("default-policy was not specified"));

    mp_user_info -> m_access_info.m_net_access_list.clear();
    mp_user_info -> m_access_info.m_fl_default_access_policy = (m_attributes.find("default-policy") -> second == "allow");

    SetNewHandler(new NetAccessHandler(mp_server, &mp_user_info -> m_access_info));
    }
  else if (element == "QueryOptions")
    SetNewHandler(new QueryOptionsHandler(mp_server, &mp_user_info -> m_access_info));
  else
    assert(0);
  }
//--------------------------------------------------------------------
void UserHandler::endElement(const XMLCh* const uri, const XMLCh* const localname,
                             const XMLCh* const qname)
  {
  std::string element = AsString(localname);

  if (element == "User")
    return PopHandler();
  else
    assert(element == "NetAccess" || element == "QueryOptions");

  m_text.remove();
  }
//--------------------------------------------------------------------
// AccessHandler implementation:
//--------------------------------------------------------------------
AccessHandler::AccessHandler(Server* ip_server, AccessRights* iop_access_rights):
  ConfigHandlerBase(ip_server), mp_access_rights(iop_access_rights) {}
//--------------------------------------------------------------------
void AccessHandler::endElement(const XMLCh* const uri, const XMLCh* const localname,
                               const XMLCh* const qname)
  {
  m_text.trim();
  std::string element = AsString(localname);

  if (element == "AllowGroup")
    mp_access_rights -> m_access_list.push_back(AccessRightsItem(AccessRightsItem::GROUP_NAME, AsString(m_text), true));
  else if (element == "DenyGroup")
    mp_access_rights -> m_access_list.push_back(AccessRightsItem(AccessRightsItem::GROUP_NAME, AsString(m_text), false));
  else if (element == "AllowUser")
    mp_access_rights -> m_access_list.push_back(AccessRightsItem(AccessRightsItem::USER_NAME, AsString(m_text), true));
  else if (element == "DenyUser")
    mp_access_rights -> m_access_list.push_back(AccessRightsItem(AccessRightsItem::USER_NAME, AsString(m_text), false));
  else if (element == "Access")
    return PopHandler();
  else
    assert(0);

  m_text.remove();
  }
//--------------------------------------------------------------------
// NetAccessHandler implementation:
//--------------------------------------------------------------------
NetAccessHandler::NetAccessHandler(Server* ip_server, AccessInfo* iop_access_info):
  ConfigHandlerBase(ip_server), mp_access_info(iop_access_info) {}
//--------------------------------------------------------------------
void NetAccessHandler::startElement(const XMLCh* const uri, const XMLCh* const localname,
                                    const XMLCh* const qname, const xercesc::Attributes& attributes)
  {
  m_text.remove();
  std::string element = AsString(localname);

  if (element == "AllowIP" || element == "DenyIP")
    CopyAttributes(attributes);
  else
    assert(0);
  }
//--------------------------------------------------------------------
void NetAccessHandler::endElement(const XMLCh* const uri, const XMLCh* const localname,
                                  const XMLCh* const qname)
  {
  m_text.trim();
  std::string element = AsString(localname);

  if (element == "AllowIP")
    {
    if (m_attributes.find("mask") != m_attributes.end())
      mp_access_info -> m_net_access_list.push_back(NetAccess(AsString(m_text), AsString(m_attributes["mask"]), true));
    else
      mp_access_info -> m_net_access_list.push_back(NetAccess(AsString(m_text), "255.255.255.255", true));
    }
  else if (element == "DenyIP")
    {
    if (m_attributes.find("mask") != m_attributes.end())
      mp_access_info -> m_net_access_list.push_back(NetAccess(AsString(m_text), AsString(m_attributes["mask"]), false));
    else
      mp_access_info -> m_net_access_list.push_back(NetAccess(AsString(m_text), "255.255.255.255", false));
    }
  else if (element == "NetAccess")
    return PopHandler();
  else
    assert(0);

  m_text.remove();
  }
//--------------------------------------------------------------------
// QueryOptionsHandler implementation:
//--------------------------------------------------------------------
QueryOptionsHandler::QueryOptionsHandler(Server* ip_server, AccessInfo* iop_access_info):
  ConfigHandlerBase(ip_server), mp_access_info(iop_access_info) {}
//--------------------------------------------------------------------
void QueryOptionsHandler::startElement(const XMLCh* const uri, const XMLCh* const localname,
                                       const XMLCh* const qname, const xercesc::Attributes& attributes)
  {
  m_text.remove();
  std::string element = AsString(localname);

  CopyAttributes(attributes);

  if (m_attributes.find("max-time") == m_attributes.end())
    throw ServerException("WrongConfig", "max-time attribute was not specified");

  if (m_attributes.find("max-entries") == m_attributes.end())
    throw ServerException("WrongConfig", "max-entries attribute was not specified");

  time_t attr_time;
  index_type attr_entries;

  if (m_attributes["max-time"] == "unlimited")
    attr_time = std::numeric_limits<time_t>::max();
  else
    attr_time = boost::lexical_cast<time_t>(AsString(m_attributes["max-time"]));

  if (m_attributes["max-entries"] == "unlimited")
    attr_entries = std::numeric_limits<index_type>::max();
  else
    attr_entries = boost::lexical_cast<index_type>(AsString(m_attributes["max-entries"]));

  if (element == "MatchOptions")
    {
    mp_access_info -> m_query_ctrl_options.m_max_match_time = attr_time;
    mp_access_info -> m_query_ctrl_options.m_max_match_entries = attr_entries;
    }
  else if (element == "DefineOptions")
    {
    mp_access_info -> m_query_ctrl_options.m_max_define_time = attr_time;
    mp_access_info -> m_query_ctrl_options.m_max_define_entries = attr_entries;
    }
  else
    assert(0);
  }
//--------------------------------------------------------------------
void QueryOptionsHandler::endElement(const XMLCh* const uri, const XMLCh* const localname,
                                     const XMLCh* const qname)
  {
  m_text.trim();
  std::string element = AsString(localname);

  if (element == "QueryOptions")
    return PopHandler();
  else
    assert(element == "MatchOptions" || element == "DefineOptions");

  m_text.remove();
  }
//--------------------------------------------------------------------
