/*****************************************************************************
 *
 * DictD++, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#include "stdafx.h"
#include "socket.h"
#include <unicode/ustring.h>

#ifdef _WIN32
typedef int socklen_t;
#else
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <netdb.h>
#include <fcntl.h>
#endif
//--------------------------------------------------------------------
Socket::Socket(): m_sock(INVALID_SOCKET)
  {
  SetInputBufferSize(1024);
  SetOutputBufferSize(16384);

  m_out_buffer_pos = 0;

  m_fl_non_blocking = false;
  m_fl_responsible = true;
  }
//--------------------------------------------------------------------
Socket::Socket(const Socket& i_socket): m_sock(i_socket.m_sock),
	  m_fl_non_blocking(i_socket.m_fl_non_blocking),
	  m_fl_responsible(true), m_out_buffer_pos(0)
  {
  m_in_buffer.resize(i_socket.m_in_buffer.size());
  m_out_buffer.resize(i_socket.m_out_buffer.size());
  i_socket.m_fl_responsible = false;
  }
//--------------------------------------------------------------------
Socket::~Socket()
  {
  if (IsValid() && m_fl_responsible)
    Close();
  }
//--------------------------------------------------------------------
int Socket::GetErrorCode() const
  {
#ifdef _WIN32
  int err_code = WSAGetLastError();
  WSASetLastError(0);
#else
  int err_code = errno;
#endif

  return err_code;
  }
//--------------------------------------------------------------------
void Socket::Create()
  {
  m_sock = socket(AF_INET, SOCK_STREAM, 0);

  if (!IsValid())
    throw SocketException("CannotCreateSocket", "Cannot create socket");
  }
//--------------------------------------------------------------------
void Socket::Bind(unsigned port)
  {
  if (!IsValid())
    throw SocketException("SocketIsNotValid", "Attempt to bind using non-valid socket");

  sockaddr_in addr;
  memset(&addr, 0, sizeof (addr));
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = INADDR_ANY;
  addr.sin_port = htons(port);

  if (bind(m_sock, (struct sockaddr *)&addr, sizeof (addr)))
    throw SocketException("CannotBindSocket", "Cannot bind socket");
  }
//--------------------------------------------------------------------
void Socket::Listen() const
  {
  if (!IsValid())
    throw SocketException("SocketIsNotValid", "Attempt to listen using non-valid socket");

  if (listen(m_sock, SOMAXCONN))
    throw SocketException("CannotListenSocket", "Cannot listen socket");
  }
//--------------------------------------------------------------------
void Socket::Accept(Socket& new_socket) const
  {
  sockaddr_in addr;
  memset(&addr, 0, sizeof (addr));
  socklen_t addr_len = sizeof(addr);
  new_socket.m_sock = accept(m_sock, (sockaddr *)&addr, &addr_len);

  if (!new_socket.IsValid())
    throw SocketException("CannotAcceptSocket", "Cannot accept socket");
  }
//--------------------------------------------------------------------
void Socket::Send(const std::string& i_str)
  {
  return Send(i_str.c_str(), (ulong)i_str.length());
  }
//--------------------------------------------------------------------
void Socket::UnquotedSend(const char* ip_str, ulong len)
  {
  if (m_out_buffer_pos + len < m_out_buffer.size())
    {
    std::copy(ip_str, ip_str + len, m_out_buffer.begin() + m_out_buffer_pos);
    m_out_buffer_pos += len;
    }
  else
    {
    std::copy(ip_str, ip_str + m_out_buffer.size() - m_out_buffer_pos, m_out_buffer.begin() + m_out_buffer_pos);

    ulong send_status = send(m_sock, &m_out_buffer[0], (ulong)m_out_buffer.size(), 0);
    len -= (ulong)m_out_buffer.size() - m_out_buffer_pos;
    ip_str += m_out_buffer.size() - m_out_buffer_pos;
    m_out_buffer_pos = 0;

    if (send_status == SOCKET_ERROR)
      throw SocketException("CannotSendData", "Cannot send data through socket");

    if (send_status < m_out_buffer.size() && !m_fl_non_blocking)
      throw SocketException("DataSendFailure", "Sent data size is smaller than size()");

    if (len >= m_out_buffer.size())
      {
      ulong send_status = send(m_sock, ip_str, len, 0);

      if (send_status == SOCKET_ERROR)
	throw SocketException("CannotSendData", "Cannot send data through socket");

      if (send_status < len && !m_fl_non_blocking)
	throw SocketException("DataSendFailure", "Sent data size is smaller than size()");
      }
    else
      {
      std::copy(ip_str, ip_str + len, m_out_buffer.begin());
      m_out_buffer_pos = len;
      }
    }
  }
//--------------------------------------------------------------------
void Socket::FlushOutput()
  {
  ulong send_status = send(m_sock, &m_out_buffer[0], m_out_buffer_pos, 0);
  m_out_buffer_pos = 0;

  if (send_status == SOCKET_ERROR)
    throw SocketException("CannotSendData", "Cannot send data through socket");

  if (send_status < m_out_buffer_pos && !m_fl_non_blocking)
    throw SocketException("DataSendFailure", "Sent data size is smaller than size()");
  }
//--------------------------------------------------------------------
void Socket::Send(const char* ip_str, ulong len)
  {
  if (!len)
    len = (ulong)strlen(ip_str);

  // We may not send lines longer than 1024 characters. Actually,
  // we may send lines up to 1024 unicode characters - this means
  // up to 1024*6 characters in UTF-8 encoding, but let's be conservative ...
  char last_char = 0;
  ulong end_sent_pos = 0, last_wsp = std::numeric_limits<ulong>::max();

  for (ulong i = 0; i < len;)
    {
    if (ip_str[i] == '\n' && last_char != '\r') // we have CR instead of CRLF
      {
      std::string str(ip_str + end_sent_pos, i - end_sent_pos);
      str.append("\r\n");

      UnquotedSend(str.c_str(), (ulong)str.size());

      last_wsp = std::numeric_limits<ulong>::max();
      end_sent_pos = ++i;
      last_char = '\n';

      continue;
      }
    else if (ip_str[i] == '.' && last_char == '\n')
      {
      std::string str(ip_str + end_sent_pos, i - end_sent_pos);
      str.append("..");

      UnquotedSend(str.c_str(), (ulong)str.size());

      last_wsp = std::numeric_limits<ulong>::max();
      end_sent_pos = ++i;
      last_char = '.';

      continue;
      }
    else if (ip_str[i] == ' ' || ip_str[i] == '\t')
      last_wsp = i;

    if (i - end_sent_pos >= 1023)
      {
      if (last_wsp != std::numeric_limits<ulong>::max())
	      {
	      std::string str(ip_str + end_sent_pos, last_wsp - end_sent_pos);
	      str.append("\r\n");

	      UnquotedSend(str.c_str(), (ulong)str.size());
	      end_sent_pos = last_wsp + 1;
        if (i == last_wsp)
          ++i;
	      last_wsp = std::numeric_limits<ulong>::max();

	      continue;
	      }
      else
	      {
	      std::string str(ip_str + end_sent_pos, i - end_sent_pos);
	      str.append("\r\n");

	      UnquotedSend(str.c_str(), (ulong)str.size());
	      end_sent_pos = i;
	      last_wsp = std::numeric_limits<ulong>::max();

	      continue;
	      }
      }

    // OK, this is usual character:
    last_char = ip_str[i++];
    }

  if (end_sent_pos != len)
    UnquotedSend(ip_str + end_sent_pos, len - end_sent_pos);
  }
//--------------------------------------------------------------------
void Socket::Receive(std::string& s)
  {
  s.clear();

  if (m_fl_non_blocking && !GetUnreadDataSize())
    return;

  do
    {
    int status = recv(m_sock, &m_in_buffer[0], (int)m_in_buffer.size(), 0);

    if (status == SOCKET_ERROR)
      throw SocketException("CannotReceiveData", "Cannot receive data through socket");

    if (status == 0)
      throw SocketException("SocketIsClosed", "Socket is closed");

    s.append(&m_in_buffer[0], status);
    }
  while (GetUnreadDataSize());
  }
//--------------------------------------------------------------------
void Socket::Connect(const std::string& host, unsigned port)
  {
  if (!IsValid())
    throw SocketException("SocketIsNotValid", "Attempt to connect using non-valid socket");

  sockaddr_in addr;
  memset(&addr, 0, sizeof (addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);

  struct hostent* he = gethostbyname(host.c_str());

  if (!he)
    {
    unsigned long status = inet_addr(host.c_str());

    if (status == INADDR_NONE)
      throw SocketException("CannotResolveName", std::string("Cannot resolve host name ") + host);

    addr.sin_addr = *(struct in_addr*)&status;
    }
  else
    addr.sin_addr = *(struct in_addr*)&he -> h_addr_list[0];

  if (connect(m_sock, (sockaddr *) &addr, sizeof(addr)))
    throw SocketException("CannotConnectSocket", "Cannot connect socket");
  }
//--------------------------------------------------------------------
bool Socket::IsValid() const
  {
  return m_sock != INVALID_SOCKET;
  }
//--------------------------------------------------------------------
unsigned long Socket::GetUnreadDataSize() const
  {
  if (!IsValid())
    throw SocketException("SocketIsNotValid", "Attempt to get unread data size using non-valid socket");

  int err_code;
  unsigned long size;

#ifdef _WIN32
  err_code = ioctlsocket(m_sock, FIONREAD, &size);
#else
  err_code = ioctl(m_sock, FIONREAD, &size);
#endif

  if (err_code)
    throw SocketException("CannotGetUnreadDataSize", "Cannot get unread data size");

  return size;
  }
//--------------------------------------------------------------------
void Socket::SetNonBlocking(bool i_fl_non_blocking)
  {
  unsigned long in = i_fl_non_blocking;

  if (i_fl_non_blocking == m_fl_non_blocking)
    return;

#ifdef _WIN32
/*
  if (WSAIoctl(m_sock, FIONBIO, &in, sizeof(unsigned long),
			  NULL, 0, NULL, NULL, NULL))
    throw SocketException("CannotSetNonBlockingMode", "Cannot set non-blocking mode");
/**/
  if (ioctlsocket(m_sock, FIONBIO, &in) == SOCKET_ERROR)
    throw SocketException("CannotSetNonBlockingMode", "Cannot set non-blocking mode");

  m_fl_non_blocking = i_fl_non_blocking;
#else
  int opts = fcntl(m_sock, F_GETFL);

  if ( opts < 0 )
    throw SocketException("CannotGetSocketFlags", "Cannot get socket flags");

  if (i_fl_non_blocking)
    opts = (opts | O_NONBLOCK);
  else
    opts = (opts & ~O_NONBLOCK);

  fcntl(m_sock, F_SETFL, opts);

  m_fl_non_blocking = i_fl_non_blocking;
#endif
  }
//--------------------------------------------------------------------
void Socket::Close()
  {
  assert(m_fl_responsible);

  int res;
#ifdef _WIN32
  res = closesocket(m_sock);
#else
  res = close(m_sock);
#endif  

  if (res)
    throw SocketException("CannotCloseSocket", "Cannot close socket");

  m_sock = INVALID_SOCKET;
  }
//--------------------------------------------------------------------
int Socket::Select(ulong i_events, long i_timeout) const
  {
  fd_set read_set, write_set, error_set;
  fd_set *p_read = NULL, *p_write = NULL, *p_error = NULL;

  if (i_events & SELECT_READ)
    {
    p_read = &read_set;
    FD_ZERO(p_read);
    FD_SET(m_sock, p_read);
    }
  if (i_events & SELECT_WRITE)
    {
    p_write = &write_set;
    FD_ZERO(p_write);
    FD_SET(m_sock, p_write);
    }
  if (i_events & SELECT_ERROR)
    {
    p_error = &error_set;
    FD_ZERO(p_error);
    FD_SET(m_sock, p_error);
    }

  int res;

  if (i_timeout == -1)
    res = select(FD_SETSIZE, p_read, p_write, p_error, NULL);
  else
    {
    timeval timeout;
    timeout.tv_sec = i_timeout;
    timeout.tv_usec = 0;

    res = select(FD_SETSIZE, p_read, p_write, p_error, &timeout);
    }

  if (!res || res == SOCKET_ERROR)
    return res;

  res = 0;

  if (p_read && FD_ISSET(m_sock, p_read))
    res |= SELECT_READ;

  if (p_write && FD_ISSET(m_sock, p_write))
    res |= SELECT_WRITE;

  if (p_error && FD_ISSET(m_sock, p_error))
    res |= SELECT_ERROR;

  return res;
  }
//--------------------------------------------------------------------
Socket& Socket::operator << (const std::string& i_str)
  {
  Socket::Send(i_str.c_str());
  return *this;
  }
//--------------------------------------------------------------------
Socket& Socket::operator << (const char* ip_str)
  {
  Socket::Send(ip_str);
  return *this;
  }
//--------------------------------------------------------------------
Socket& Socket::operator << (const UnicodeString& s)
  {
  char* p_buf = new char[s.length()*6 + 1];
  int32_t len;

  UErrorCode status = U_ZERO_ERROR;
  u_strToUTF8(p_buf, s.length()*6 + 1, &len, s.getBuffer(), s.length(), &status);

  assert(len < s.length()*6 + 1);

  Socket::Send(p_buf, len);

  delete[] p_buf;

  return *this;
  }
//--------------------------------------------------------------------
Socket& Socket::operator >> (std::string& s)
  {
  Socket::Receive(s);
  return *this;
  }
//--------------------------------------------------------------------
uint32_t Socket::GetPeerAddress() const
  {
  sockaddr_in addr;
  socklen_t len = sizeof(addr);
  if (getpeername(m_sock, (sockaddr*)&addr, &len) || addr.sin_family != AF_INET ||
      sizeof(addr) != len)
    throw SocketException("CannotGetPeerName", "getpeername() call failed");

#ifdef _WIN32
  return addr.sin_addr.S_un.S_addr;
#else
  return addr.sin_addr.s_addr;
#endif
  }
//--------------------------------------------------------------------
